import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec
import seaborn as sn
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.colors as colors
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from concurrent.futures import ProcessPoolExecutor as PPE
from concurrent.futures import as_completed
from skimage import measure
from time import time
from tqdm import tqdm
from pickle import loads
#import pandas as pd
import vtk
import plotly.graph_objects as go

from ..tests.bumpy_sphere import bumpy_sphere

def mpu_meshgrid(mpu_object,buf=1.25,res=10,workers=1,k=None,plane_axis=None,plane_value=None,marching=False,gradient=False,record=False,global_f=False):
    dims = []
    if not marching:
        for d in range(mpu_object.ddim):
            if d is plane_axis:
                continue
            dims.append(np.linspace(mpu_object.dim_range[d*2]-buf,
                                    mpu_object.dim_range[d*2+1]+buf,res))
    else:
        res = res*1j
        x,y,z = np.ogrid[mpu_object.dim_range[0]-buf:mpu_object.dim_range[1]+buf:res,
                         mpu_object.dim_range[2]-buf:mpu_object.dim_range[3]+buf:res,
                         mpu_object.dim_range[4]-buf:mpu_object.dim_range[5]+buf:res]
        results = mpu_object.function_marching(x,y,z)
        return (x,y,z),results
    DIMS = np.meshgrid(*dims)
    DIMSf = []
    for d in range(len(DIMS)):
        DIMSf.append(DIMS[d].flatten())
    value = []
    if plane_value is not None:
        plane_values = [plane_value]*len(DIMSf[0])
        DIMSf.insert(plane_axis,plane_values)
        plane = np.array(plane_values).reshape(DIMS[0].shape)
        DIMS.insert(plane_axis,plane)
    if workers > 1:
        chunksize = len(DIMSf[0])//workers
        number_of_chunks = len(DIMSf[0])//chunksize
        executor = PPE(max_workers=workers)
        if global_f:
            pass
        elif k is not None:
            k = [k]*len(DIMSf[0])
            DIMSf.append(k)
        else:
            k = [len(mpu_object.patches)]*len(DIMSf[0])
            DIMSf.append(k)
        if not gradient:
            result_generator  = executor.map(loads(mpu_object.pickled_DD[0]),zip(*DIMSf),chunksize=chunksize)
        else:
            result_generator  = executor.map(loads(mpu_object.pickled_DD[1]),zip(*DIMSf),chunksize=chunksize)
        executor.shutdown(wait=True)
        if not gradient:
            results = []
            for result in result_generator:
                results.append(result)
            results = np.array(results)
            results = results.reshape(DIMS[0].shape)
        else:
            u = []
            v = []
            w = []
            for result in result_generator:
                u.append(result[0])
                v.append(result[1])
                w.append(result[2])
            u = np.array(u)
            u = u.reshape(DIMS[0].shape)
            v = np.array(v)
            v = u.reshape(DIMS[0].shape)
            w = np.array(w)
            w = u.reshape(DIMS[0].shape)
            results = [u,v,w]
    else:
        time_data = []
        results = []
        if global_f:
            pass
        elif k is not None:
            k = [k]*len(DIMSf[0])
            DIMSf.append(k)
        else:
            k = [len(mpu_object.patches)]*len(DIMSf[0])
            DIMSf.append(k)
        if not gradient:
            for i in zip(*DIMSf):
                start = time()
                results.append(mpu_object.DD[0](i))
                stop = time() - start
                time_data.append(stop)
            results = np.array(results)
            results = results.reshape(DIMS[0].shape)
            if record:
                results = time_data
        else:
            u = []
            v = []
            w = []
            for i in zip(*DIMSf):
                start = time()
                results = mpu_object.DD[1](i)
                stop = time() - start
                time_data.append(stop)
                u.append(results[0])
                v.append(results[1])
                w.append(results[2])
            u = np.array(u)
            u = u.reshape(DIMS[0].shape)
            v = np.array(v)
            v = v.reshape(DIMS[0].shape)
            w = np.array(w)
            w = w.reshape(DIMS[0].shape)
            results = [u,v,w]
            if record:
                results = time_data
    return DIMS,results

def plot_volume(mpu_object,resolution=20,workers=1,cmin=-1,cmax=0,surface_count=14,k=None,global_f=False):
    DIMS,results = mpu_meshgrid(mpu_object,res=resolution,workers=workers,k=k,global_f=global_f)
    fig = go.Figure(data=(go.Isosurface(
                    x=DIMS[0].flatten(),
                    y=DIMS[1].flatten(),
                    z=DIMS[2].flatten(),
                    value=results.flatten(),
                    opacity=0.1,
                    isomin=cmin,
                    isomax=cmax,
                    surface_count=surface_count),go.Scatter3d(x = mpu_object.points[:,0],y=mpu_object.points[:,1],z=mpu_object.points[:,2])))
    fig.show()

def plot_volume_individual(func,x_range,y_range,z_range,buf=1.25,res=20,surface_count=15,cmin=-1,cmax=0):
    res = res*1j
    x,y,z = np.ogrid[x_range[0]-buf:x_range[1]+buf:res,
                     y_range[0]-buf:y_range[1]+buf:res,
                     z_range[0]-buf:z_range[1]+buf:res]
    results = func(x,y,z)
    fig = go.Figure(data=(go.Isosurface(
                    x=x.flatten(),
                    y=y.flatten(),
                    z=z.flatten(),
                    value=results.flatten(),
                    opacity=0.1,
                    isomin=cmin,
                    isomax=cmax,
                    surface_count=surface_count)))
    fig.show()

def plot_slice_error(mpu_object,resolution=20,k=None,workers=1,plane_axis=2,plane_value=0,unit_scale='mm'):
    DIMS,results_exact = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                      k=len(mpu_object.patches),plane_axis=plane_axis,
                                      plane_value=plane_value)
    DIMS,results_approximate = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                      k=k,plane_axis=plane_axis,plane_value=plane_value)
    DIMSv,voxels = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                k=len(mpu_object.patches),marching=True)
    verts, faces, normals, values = measure.marching_cubes(voxels,level=0)
    verts *= np.array([np.diff(ar.flat)[0] for ar in [DIMSv[0],DIMSv[1],DIMSv[2]]])
    verts += np.array([DIMSv[0].min(),DIMSv[1].min(),DIMSv[2].min()])

    error = results_exact - results_approximate
    error_min = error.min()
    error_max = error.max()
    error_divide = max(abs(error_min),abs(error_max))
    error = error/(error_divide)
    colormap = plt.cm.RdBu
    main_axes = list(range(len(DIMS)))
    main_axes.remove(plane_axis)
    axes_labels = ['x ({})'.format(unit_scale),'y ({})'.format(unit_scale),'z ({})'.format(unit_scale)]
    axes_labels_all = ['x ({})'.format(unit_scale),'y ({})'.format(unit_scale),'z ({})'.format(unit_scale)]
    axes_labels.remove(axes_labels[plane_axis])
    fig = plt.figure(figsize=(12,6))
    spec = gridspec.GridSpec(ncols=4, nrows=1,
                             width_ratios=[1, 1, 1, 1.5])
    ax1 = fig.add_subplot(spec[0])
    ax2 = fig.add_subplot(spec[1])
    ax3 = fig.add_subplot(spec[2])
    ax4 = fig.add_subplot(spec[3],projection='3d')

    cax1 = ax1.contourf(DIMS[main_axes[0]],DIMS[main_axes[1]],results_exact,norm=colors.TwoSlopeNorm(0))
    ax1.contour(DIMS[main_axes[0]],DIMS[main_axes[1]],results_exact,levels=[0],linewidths=(2,))
    #ax1.set_title('Exact Interpolation \n {} Patches'.format(len(mpu_object.patches)))
    ax1.set_xlabel(axes_labels[0])
    ax1.set_ylabel(axes_labels[1])

    cax2 = ax2.contourf(DIMS[main_axes[0]],DIMS[main_axes[1]],results_approximate,norm=colors.TwoSlopeNorm(0))
    ax2.contour(DIMS[main_axes[0]],DIMS[main_axes[1]],results_approximate,levels=[0],linewidths=(2,))
    #ax2.set_title('Approximate Interpolation \n {} Patches'.format(k))
    ax2.set_xlabel(axes_labels[0])

    cax3 = ax3.contourf(DIMS[main_axes[0]],DIMS[main_axes[1]],error,cmap=colormap,norm=colors.TwoSlopeNorm(0))

    #ax3.set_title('Relative Truncation Error')
    ax3.set_xlabel(axes_labels[0])

    mesh = Poly3DCollection(verts[faces])
    mesh.set_edgecolor('k')
    ax4.add_collection3d(mesh)
    ax4.plot_surface(DIMS[0],DIMS[1],DIMS[2],alpha=0.2,color='r')
    ax4.set_xlim(DIMSv[0].min(),DIMSv[0].max())
    ax4.set_ylim(DIMSv[1].min(),DIMSv[1].max())
    ax4.set_zlim(DIMSv[2].min(),DIMSv[2].max())
    ax4.set_xlabel(axes_labels_all[0])
    ax4.set_ylabel(axes_labels_all[1])
    ax4.set_zlabel(axes_labels_all[2])
    plt.draw()
    p0 = ax1.get_position().get_points().flatten()
    p1 = ax2.get_position().get_points().flatten()
    p2 = ax3.get_position().get_points().flatten()
    ax_cbar = fig.add_axes([p0[0], 0.1, p1[2]-p0[0], 0.05])
    ax_cbar.set_in_layout(True)
    cbar1 = plt.colorbar(cax1, cax=ax_cbar, orientation='horizontal', pad=0.25)
    labels = cbar1.ax.get_xticklabels()
    cbar1.ax.set_xticklabels(labels, rotation=40)


    ax_cbar1 = fig.add_axes([p2[0], 0.1, p2[2]-p2[0], 0.05])
    ax_cbar1.set_in_layout(True)
    cbar2 = plt.colorbar(cax3, cax=ax_cbar1, orientation='horizontal', pad=0.25)
    labels = cbar2.ax.get_xticklabels()
    cbar2.ax.set_xticklabels(labels, rotation=40)
    plt.subplots_adjust(bottom=0.25)

    plt.show()

def level_contour(mpu_object,resolution=20,k=2,buf=1.25,level=0,workers=1,plane_axis=2,plane_value=0.5,unit_scale='mm',name="untitled"):
    if plane_axis == 0:
        plane_value = (mpu_object.x_range[1]-mpu_object.x_range[0])*plane_value + mpu_object.x_range[0]
    elif plane_axis == 1:
        plane_value = (mpu_object.y_range[1]-mpu_object.y_range[0])*plane_value + mpu_object.y_range[0]
    elif plane_axis == 2:
        plane_value = (mpu_object.z_range[1]-mpu_object.z_range[0])*plane_value + mpu_object.z_range[0]
    DIMS,results_exact = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                      k=k,plane_axis=plane_axis,
                                      plane_value=plane_value,buf=buf)
    colormap = plt.cm.RdBu
    main_axes = list(range(len(DIMS)))
    main_axes.remove(plane_axis)
    fig,ax = plt.subplots()
    CS = ax.contour(DIMS[main_axes[0]],DIMS[main_axes[1]],results_exact,levels=[level],linewidths=(2,))
    #CS.collections[0].set_label(r"$\Gamma_{" +str(len(mpu_object.patches))+ r"}$")
    return CS

def plot_mpu_boundaries(mpu_object,resolution=20,k=[2],buf=1.25,level=0,workers=1,plane_axis=2,plane_value=0.5,test_size=100,unit_scale='mm',name="untitled"):
    if plane_axis == 0:
        plane_value = (mpu_object.x_range[1]-mpu_object.x_range[0])*plane_value + mpu_object.x_range[0]
    elif plane_axis == 1:
        plane_value = (mpu_object.y_range[1]-mpu_object.y_range[0])*plane_value + mpu_object.y_range[0]
    elif plane_axis == 2:
        plane_value = (mpu_object.z_range[1]-mpu_object.z_range[0])*plane_value + mpu_object.z_range[0]
    DIMS,results_exact = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                      k=len(mpu_object.patches),plane_axis=plane_axis,
                                      plane_value=plane_value,buf=buf)
    APPROX = []
    for ki in k:
        DIMS,approximate = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                          k=ki,plane_axis=plane_axis,plane_value=plane_value)
        APPROX.append(approximate)

    colormap = plt.cm.RdBu
    main_axes = list(range(len(DIMS)))
    main_axes.remove(plane_axis)
    axes_labels = ['x ({})'.format(unit_scale),'y ({})'.format(unit_scale),'z ({})'.format(unit_scale)]
    axes_labels_all = ['x ({})'.format(unit_scale),'y ({})'.format(unit_scale),'z ({})'.format(unit_scale)]
    axes_labels.remove(axes_labels[plane_axis])
    fig,ax = plt.subplots()
    #spec = gridspec.GridSpec(ncols=4, nrows=1,
    #                         width_ratios=[1, 1, 1, 1.5])
    #ax1 = fig.add_subplot(spec[0])
    #ax2 = fig.add_subplot(spec[1])
    #ax3 = fig.add_subplot(spec[2])
    #ax4 = fig.add_subplot(spec[3],projection='3d')
    positions = []
    cax = ax.contourf(DIMS[main_axes[0]],DIMS[main_axes[1]],results_exact,norm=colors.TwoSlopeNorm(0),cmap="RdBu_r",alpha=0.5)
    CS = ax.contour(DIMS[main_axes[0]],DIMS[main_axes[1]],results_exact,levels=[level],linewidths=(2,))
    CS.collections[0].set_label(r"$\Gamma_{" +str(len(mpu_object.patches))+ r"}$")
    #fmt = {CS.levels[0]:r"$\Gamma_{" +str(len(mpu_object.patches)) + r"}$"}
    #ax.clabel(CS,fmt=fmt,fontsize=10)
    for c,ap in enumerate(APPROX):
        tmp = ax.contour(DIMS[main_axes[0]],DIMS[main_axes[1]],ap,levels=[level],linewidths=(2,),linestyles='--')
        tmp.collections[0].set_label(r"$\Gamma_{" +str(k[c])+ r"}$")
        #fmt = {tmp.levels[0]:r"$\Gamma_{" +str(k[c])+ r"}$"}
        #ax.clabel(tmp,fmt=fmt,fontsize=10)
    plt.legend()
    pts = np.random.random((test_size,3))
    xr = (np.min(DIMS[0].flatten()),np.max(DIMS[0].flatten()))
    yr = (np.min(DIMS[1].flatten()),np.max(DIMS[1].flatten()))
    zr = (np.min(DIMS[2].flatten()),np.max(DIMS[2].flatten()))
    pts[:,0] = (pts[:,0]*2-1)*(xr[1] - xr[0])/2 + (mpu_object.x_range[0] + mpu_object.x_range[1])/2
    pts[:,1] = (pts[:,1]*2-1)*(yr[1] - yr[0])/2 + (mpu_object.y_range[0] + mpu_object.y_range[1])/2
    pts[:,2] = (pts[:,2]*2-1)*(zr[1] - zr[0])/2 + (mpu_object.z_range[0] + mpu_object.z_range[1])/2
    pts[:,plane_axis] = plane_value
    interior = []
    exterior = []
    interior_wrong = []
    exterior_wrong = []
    for i in range(pts.shape[0]):
        if mpu_object.within(pts[i,0],pts[i,1],pts[i,2],len(mpu_object.patches),level=level)[0]:
            if mpu_object.within(pts[i,0],pts[i,1],pts[i,2],k[0],level=level)[0]:
                interior.append(pts[i,:].tolist())
            else:
                exterior_wrong.append(pts[i,:].tolist())
        else:
            if not mpu_object.within(pts[i,0],pts[i,1],pts[i,2],k[0],level=level)[0]:
                exterior.append(pts[i,:].tolist())
            else:
                interior_wrong.append(pts[i,:].tolist())
    interior = np.array(interior)
    exterior = np.array(exterior)
    interior_wrong = np.array(interior_wrong)
    exterior_wrong = np.array(exterior_wrong)
    if interior.shape[0] > 0:
        if interior.shape[0] > 100:
            stop = 99
        else:
            stop = None
        ax.scatter(interior[:stop,main_axes[0]],interior[:stop,main_axes[1]],c='blue',alpha=0.7)
    if exterior.shape[0] > 0:
        if exterior.shape[0] > 100:
            stop = 99
        else:
            stop = None
        ax.scatter(exterior[:stop,main_axes[0]],exterior[:stop,main_axes[1]],c='red',alpha=0.7)
    if exterior_wrong.shape[0] > 0:
        if exterior_wrong.shape[0] > 100:
            stop = 99
        else:
            stop = None
        ax.scatter(exterior_wrong[:stop,main_axes[0]],exterior_wrong[:stop,main_axes[1]],c='red',alpha=0.7,marker='x')
    if interior_wrong.shape[0] > 0:
        if interior_wrong.shape[0] > 100:
            stop = 99
        else:
            stop = None
        ax.scatter(interior_wrong[:stop,main_axes[0]],interior_wrong[:stop,main_axes[1]],c='blue',alpha=0.7,marker='x')
    plt.savefig(name+'.svg', format="svg", bbox_inches='tight')
    return interior.shape[0],exterior.shape[0],interior_wrong.shape[0],exterior_wrong.shape[0],test_size

def plot_gradient(mpu_object,workers=1,k=None,resolution=20,plane_axis=2,plane_value=0,
                  contour_overlay=True,unit_scale='mm'):
    DIMS,gradient = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                 plane_axis=plane_axis,plane_value=plane_value,
                                 gradient=True)
    DIMS,gradient_approximate = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                             plane_axis=plane_axis,plane_value=plane_value,
                                             k=k,gradient=True)
    if contour_overlay:
        DIMS_c,result = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                     plane_axis=plane_axis,plane_value=plane_value)
    main_axes = list(range(len(DIMS)))
    main_axes.remove(main_axes[plane_axis]) #use del?
    fig = plt.figure(figsize=(6,5))
    ax = fig.add_subplot(111)

    if contour_overlay:
        cax = ax.contourf(DIMS[main_axes[0]],DIMS[main_axes[1]],result,norm=colors.TwoSlopeNorm(0))
        p0 = ax.get_position().get_points().flatten()
        ax_cbar = fig.add_axes([p0[0], 0.1, p0[2]-p0[0], 0.05])
        ax_cbar.set_in_layout(True)
        cbar1 = plt.colorbar(cax, cax=ax_cbar, orientation='horizontal', pad=0.25)
        labels = cbar1.ax.get_xticklabels()
        cbar1.ax.set_xticklabels(labels, rotation=40)

    ax.quiver(DIMS[main_axes[0]],DIMS[main_axes[1]],
              gradient[main_axes[0]],gradient[main_axes[1]])

    ax.quiver(DIMS[main_axes[0]],DIMS[main_axes[1]],
              gradient_approximate[main_axes[0]],gradient_approximate[main_axes[1]],color='r')
    axes_labels = ['x ({})'.format(unit_scale),'y ({})'.format(unit_scale),'z ({})'.format(unit_scale)]
    axes_labels.remove(axes_labels[plane_axis])
    ax.set_xlabel(axes_labels[0])
    ax.set_ylabel(axes_labels[1])
    plt.subplots_adjust(bottom=0.25)
    plt.show()

def plot_gradient_error(mpu_object,workers=1,k=None,resolution=20,plane_axis=2,plane_value=0,
                  contour_overlay=True,unit_scale='mm'):
    DIMS,gradient = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                 plane_axis=plane_axis,plane_value=plane_value,
                                 gradient=True)
    DIMS,approximate = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                    plane_axis=plane_axis,plane_value=plane_value,
                                    k=k,gradient=True)
    DIMS_c,result = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                 plane_axis=plane_axis,plane_value=plane_value)
    gradient = [gradient[0].flatten(),gradient[1].flatten(),gradient[2].flatten()]
    approximate = [approximate[0].flatten(),approximate[1].flatten(),approximate[2].flatten()]
    alignment_error = []
    magnitude_error = []
    for i in range(len(gradient[0])):
        grad_mag = (gradient[0][i]**2+gradient[1][i]**2+gradient[2][i]**2)**(1/2)
        approx_mag = (approximate[0][i]**2+approximate[1][i]**2+approximate[2][i]**2)**(1/2)
        alignment_error.append((gradient[0][i]/grad_mag)*(approximate[0][i]/approx_mag)+(gradient[1][i]/grad_mag)*(approximate[1][i]/approx_mag)+
                               (gradient[2][i]/grad_mag)*(approximate[2][i]/approx_mag))
        magnitude_error.append(grad_mag-approx_mag)
    fig = plt.figure()
    ax1 = fig.add_subplot(121)
    ax1.scatter(result.flatten(),alignment_error)
    ax1.set_xlabel('Contour Layer')
    ax1.set_ylabel('Alignment Difference')
    ax2 = fig.add_subplot(122)
    ax2.scatter(result.flatten(),magnitude_error)
    ax2.set_xlabel('Contour Layer')
    ax2.set_ylabel('Magnitude Difference')
    plt.show()

def plot_error(mpu_object,workers=1,k=None,resolution=20,plane_axis=2,plane_value=0):
    DIMS,result = mpu_meshgrid(mpu_object,workers=workers,res=resolution,plane_axis=plane_axis,
                               plane_value=plane_value)
    DIMS,approx = mpu_meshgrid(mpu_object,workers=workers,k=k,res=resolution,plane_axis=plane_axis,
                               plane_value=plane_value)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    main_axes = list(range(len(DIMS)))
    main_axes.remove(main_axes[plane_axis])
    contours = ax.contour(DIMS[main_axes[0]],DIMS[main_axes[1]],result)
    levels = contours.levels[1:-1]
    level_areas = []
    level_error = []
    error_per_area = []
    resultf = result.flatten()
    approxf = approx.flatten()
    l0 = -1
    a0 = 0
    passed_levels = []
    for idx, level in enumerate(levels):
        #contour = contours.collections[idx]
        #if len(contour.get_paths()) == 0:
        #    continue
        #else:
        #    passed_levels.append(level)
        #vs = contour.get_paths()[0].vertices
        #a = 0
        #d0_0,d1_0 = vs[0]
        #for [d0_1,d1_1] in vs[1:]:
        #    dd0 = d0_1 - d0_0
        #    dd1 = d1_1 - d1_0
        #    a += 0.5*(d1_0*dd0 - d0_0*dd1)
        #    d0_0 = d0_1
        #    d1_0 = d1_1
        #level_areas.append(a-a0)
        #a0 = a
        lower_filter = set(np.argwhere(l0<resultf).flatten())
        upper_filter = set(np.argwhere(resultf<level).flatten())
        bounded_filter = list(lower_filter.intersection(upper_filter))
        level_error.append(np.mean(resultf[bounded_filter] - approxf[bounded_filter]))
        l0 = level
        #error_per_area.append(level_error[-1]/level_areas[-1])
    plt.clf()
    plt.plot(levels,level_error)
    plt.show()

def plot_time(mpu_object,resolution=20,k_start=2,k_end=100,gradient=False,plane_axis=None,plane_value=None):
    all_time = []
    mean_time = []
    k_values  = []
    sd_time_values = []
    all_error = []
    mean_error = []
    sd_error = []
    all_k = []
    #d_type = []
    DIMS,exact = mpu_meshgrid(mpu_object,k=None,res=resolution,gradient=gradient,
                                  plane_axis=plane_axis,plane_value=plane_value)
    for i in tqdm(range(k_start,k_end+1),desc='Evaluating Time Grids  '):
        k_values.append(i)
        DIMS,time_data = mpu_meshgrid(mpu_object,k=i,res=resolution,gradient=gradient,
                                      plane_axis=plane_axis,plane_value=plane_value,
                                      record=True)
        DIMS,approx = mpu_meshgrid(mpu_object,k=i,res=resolution,gradient=gradient,
                                      plane_axis=plane_axis,plane_value=plane_value)
        error = abs(exact.flatten() - approx.flatten())
        k_tmp = [i]*len(error)
        #d_type.extend(['error']*len(error))
        #d_type.extend(['time']*len(time_data))
        mean_error.append(np.mean(error))
        sd_error.append(np.std(error))
        all_time.extend(time_data)
        mean_time.append(np.mean(time_data))
        sd_time_values.append(np.std(time_data))
        all_error.extend(error)
        all_k.extend(k_tmp)

    data = pd.DataFrame()
    data['times'] = all_time
    data['k'] = all_k
    data['error'] = all_error
    sn.lineplot(data=data,x='k',y='times')
    ax2 = plt.twinx()
    sn.lineplot(data=data,x='k',y='error',ax=ax2,color='red')
    plt.show()

def time_PU_global(mpu_object):
    start = time()
    mpu_object.solve()
    pu_time = time()-start
    start = time()
    mpu_object.solve(PU=False)
    global_time = time()-start
    return pu_time,global_time

#def pu_v_global(mpu_object,start=10,stop=40):

def show_path(points,points2):
    fig = plt.figure()
    ax  = fig.add_subplot(111,projection='3d')
    ax.scatter3D(points[0,0],points[0,1],points[0,2],color='g')
    ax.scatter3D(points[1:,0],points[1:,1],points[1:,2],color='r')
    ax.scatter3D(points2[:,0],points2[:,1],points2[:,2],color='b')
    plt.show()

def show_mesh(points,points2,points3):
    fig = plt.figure()
    ax  = fig.add_subplot(111,projection='3d')
    ax.scatter3D(points[:,0],points[:,1],points[:,2],color='r')
    ax.scatter3D(points2[:,0],points2[:,1],points2[:,2],color='b')
    ax.scatter3D(points3[:,0],points3[:,1],points3[:,2],color='g')
    plt.show()

def newton(mpu_object):
    f1 = mpu_object.function([1,1,1,None])
    f2 = mpu_object.function([1.1,1,1,None])
    g = np.array(mpu_object.gradient([1,1,1,None]))
    print('F1: {}'.format(f1))
    print('F2: {}'.format(f2))
    print('Ap: {}'.format(f1+np.dot(np.array([0.1,0,0]),g)))

def show_patches(mpu_object):
    patch_sets = [set(tuple([tuple(pp) for pp in p.points.tolist()])) for p in mpu_object.patches]
    unique_points = []
    overlap_points = []
    for patch_idx,patch_i in enumerate(patch_sets):
        patch_i_diff = patch_i.difference({})
        patch_i_overlap = set([])
        for patch_jdx,patch_j in enumerate(patch_sets):
            if patch_idx == patch_jdx:
                continue
            else:
                patch_i_diff = patch_i_diff.difference(patch_j)
                patch_i_overlap.update(patch_i.intersection(patch_j))
            print(len(unique_points))
        unique_points.extend(list(patch_i_diff))
        overlap_points.extend(list(patch_i_overlap))
    unique_points = np.array(unique_points)
    overlap_points = np.array(overlap_points)
    fig = plt.figure()
    ax  = fig.add_subplot(111,projection='3d')
    ax.scatter3D(unique_points[:,0],unique_points[:,1],unique_points[:,2],color='r')
    ax.scatter3D(overlap_points[:,0],overlap_points[:,1],overlap_points[:,2],color='b')
    plt.show()
#get area of contourf for L1 error estimate as a function of "tissue depth" which will basically argue
#that for our area of interest we can ignore the problematic areas of unstable far-field growth

#for gradient error take the angle and magnitude differences and also show them as a function of tissue
# layer position



import numpy as np
import numba as nb
from functools import partial
from copy import deepcopy
from scipy.optimize import minimize,approx_fprime

@nb.jit(nopython=True)
def fast_local_function(data,mesh,terminal,edge,gamma,nu,Qterm,Pperm,Pterm):
    ####################
    reduced_resistances = []
    reduced_lengths     = []
    bifurcations        = []
    flows               = [] # should only be changing path (not alt paths)
    main_idx            = [] # index 0: data[:,15]; index 1: data[:,16]
    alt_idx             = []
    new_scale           = [] # compound bifurcations
    alt_scale           = []
    ####################
    total_flow          = data[edge,22].item()+Qterm*2
    upstream            = data[edge,0:3]
    downstream          = data[edge,3:6]
    upstream_length     = np.sqrt(np.sum(np.square(mesh-upstream),axis=1))
    downstream_length   = np.sqrt(np.sum(np.square(mesh-downstream),axis=1))
    terminal_length     = np.sqrt(np.sum(np.square(mesh-terminal),axis=1))
    # Begin calculating reduced resistance and bifurcation ratios
    R_terminal          = (8*nu/np.pi)*terminal_length
    R_terminal_sister   = (8*nu/np.pi)*downstream_length + (data[edge,25]-(8*nu/np.pi)*data[edge,20])
    f_terminal          = (1+((data[edge,22]*R_terminal_sister)/(Qterm*R_terminal))**(gamma/4)) ** (-1/gamma)
    f_terminal_sister   = (1+((data[edge,22]*R_terminal_sister)/(Qterm*R_terminal))**(-gamma/4)) ** (-1/gamma)
    R_0                 = (8*nu/np.pi)*upstream_length + ((f_terminal**4)/R_terminal +
                          (f_terminal_sister**4)/R_terminal_sister)**-1
    L_0                 = f_terminal ** 2 * terminal_length + f_terminal_sister ** 2 * (downstream_length + data[edge,27])
    reduced_resistances.append(R_0)
    reduced_lengths.append(L_0)
    flows.append(data[edge,22].item()+Qterm)
    main_idx.append(edge.item())
    bifurcations.append(f_terminal_sister) # 0 -> 15; 1-> 16
    bifurcations.append(f_terminal)
    if edge == 0:
        # on the terminating round the new terminal is always 16 and the existing is 15
        #bifurcations.append([f_terminal_sister,f_terminal])
        L_0 = upstream_length + L_0
        R_0 = (((data[edge,22]+Qterm)*R_0)/(Pperm-Pterm))**(1/4)
        alt_idx.append(-1)
        new_scale.append(np.ones(f_terminal.shape))
        alt_scale.append(np.ones(f_terminal.shape))
        return (R_0,L_0,f_terminal,f_terminal_sister,None,mesh,reduced_resistances,
                reduced_lengths,bifurcations,flows,main_idx,alt_idx,new_scale,alt_scale,R_terminal,
                R_terminal_sister)
    previous = edge
    edge = int(data[edge,17].item())
    if int(data[edge,15].item()) == previous:
        alt = int(data[edge,16].item())
        if alt == -1:
            alt = -2
        alt_idx.append(alt)
        position = 0
        #bifurcations.append([f_terminal,f_terminal_sister])
    else:
        alt = int(data[edge,15].item())
        if alt == -1:
            alt = -2
        alt_idx.append(alt)
        position = 1
        #bifurcations.append([f_terminal_sister,f_terminal])
    if alt == -2:
        f_changed = np.array([1.0])
        f_stagnant = np.array([0.0])
        f_terminal = f_terminal*f_changed
        f_terminal_sister = f_terminal_sister*f_changed
        f_parent = f_changed
        R_0 = (8*nu/np.pi)*data[edge,20].item() + R_0
        L_0 = f_changed ** 2 *(upstream_length+L_0)
    else:
        f_changed = (1+((data[alt,22]*data[alt,25])/((data[previous,22]+Qterm)*R_0)) ** (gamma/4)) ** (-1/gamma)
        f_stagnant = (1+((data[alt,22]*data[alt,25])/((data[previous,22]+Qterm)*R_0)) ** (-gamma/4)) ** (-1/gamma)
        f_terminal = f_terminal*f_changed
        f_terminal_sister = f_terminal_sister*f_changed
        f_parent = f_changed
        R_0 = (8*nu/np.pi)*data[edge,20] + ((f_changed ** 4) /R_0 + (f_stagnant ** 4) /data[alt,25]) ** -1
        L_0 = f_changed ** 2 *(upstream_length+L_0) + f_stagnant ** 2 * (data[alt,20] + data[alt,27])
    reduced_resistances.append(R_0)
    reduced_lengths.append(L_0)
    flows.append(data[edge,22].item()+Qterm)
    main_idx.append(edge)
    new_scale.append(f_changed)
    alt_scale.append(f_stagnant)
    if position == 0:
        bifurcations.append(f_changed)
        bifurcations.append(f_stagnant)
    else:
        bifurcations.append(f_stagnant)
        bifurcations.append(f_changed)
    if edge == 0:
        L_0 = data[edge,20]+L_0
        R_0 = (((data[edge,22]+Qterm)*R_0)/(Pperm-Pterm))**(1/4)
        return (R_0,L_0,f_terminal,f_terminal_sister,f_parent,mesh,reduced_resistances,
                reduced_lengths,bifurcations,flows,main_idx,alt_idx,new_scale,alt_scale,R_terminal,
                R_terminal_sister)
    previous = edge
    edge = int(data[edge,17].item())
    while edge >= 0:
        if int(data[edge, 15].item()) == previous:
            alt = int(data[edge,16].item())
            if alt == -1:
                alt = -2
            alt_idx.append(alt)
            position = 0
        else:
            alt = int(data[edge,15].item())
            if alt == -1:
                alt = -2
            alt_idx.append(alt)
            position = 1
        if alt == -2:
            f_changed = np.array([1.0])
            f_stagnant = np.array([0.0])
            R_0 =  (8*nu/np.pi)*data[edge,20] + R_0
            L_0 =  f_changed ** 2 * (data[previous, 20] + L_0)
            f_terminal = f_terminal*f_changed
            f_terminal_sister = f_terminal_sister*f_changed
            f_parent = f_parent*f_changed
        else:
            f_changed = (1+((data[alt, 22]*data[alt,25])/((Qterm+data[previous,22])*R_0)) ** (gamma/4)) ** (-1/gamma)
            f_stagnant = (1+((data[alt, 22]*data[alt,25])/((Qterm+data[previous,22])*R_0)) ** (-gamma/4)) ** (-1/gamma)
            f_terminal = f_terminal*f_changed
            f_terminal_sister = f_terminal_sister*f_changed
            f_parent = f_parent*f_changed
            R_0 = (8*nu/np.pi)*data[edge,20] + ((f_changed ** 4 / R_0) + (f_stagnant ** 4 / data[alt, 25])) ** -1
            L_0 = f_changed ** 2 * (data[previous, 20] + L_0) + f_stagnant ** 2 *(data[alt, 20] + data[alt, 27])
        reduced_resistances.append(R_0)
        reduced_lengths.append(L_0)
        flows.append(data[edge,22].item()+Qterm)
        main_idx.append(edge)
        if edge >= 0:
            for j in range(len(new_scale)):
                new_scale[j] = new_scale[j]*f_changed
                alt_scale[j] = alt_scale[j]*f_changed
            new_scale.append(f_changed)
            #alt_scale = [bif*f_changed for bif in alt_scale]
            alt_scale.append(f_stagnant)
        #else:
        #    new_scale.append(np.ones(f_changed.shape))
        #    alt_scale.append(np.ones(f_changed.shape))
        if position == 0:
            bifurcations.append(f_changed)
            bifurcations.append(f_stagnant)
        else:
            bifurcations.append(f_stagnant)
            bifurcations.append(f_changed)
        previous = edge
        edge = int(data[edge,17].item())
    L_0 = data[0,20] + L_0
    R_0 = (((data[0, 22]+Qterm)*R_0)/(Pperm-Pterm))**(1/4)
    return (R_0,L_0,f_terminal,f_terminal_sister,f_parent,mesh,reduced_resistances,
            reduced_lengths,bifurcations,flows,main_idx,alt_idx,new_scale,alt_scale,R_terminal,
            R_terminal_sister)


def fast_local_function_test(data,mesh,terminal,edge,gamma,nu,Qterm,Pperm,Pterm):
    ####################
    reduced_resistances = []
    reduced_lengths     = []
    bifurcations        = []
    flows               = [] # should only be changing path (not alt paths)
    main_idx            = [] # index 0: data[:,15]; index 1: data[:,16]
    alt_idx             = []
    new_scale           = [] # compound bifurcations
    alt_scale           = []
    ####################
    #total_flow          = data[edge,22].item()+Qterm*2
    upstream            = data[edge,0:3]
    downstream          = data[edge,3:6]
    upstream_length     = np.sqrt(np.sum(np.square(mesh-upstream),axis=1))
    downstream_length   = np.sqrt(np.sum(np.square(mesh-downstream),axis=1))
    terminal_length     = np.sqrt(np.sum(np.square(mesh-terminal),axis=1))
    # Begin calculating reduced resistance and bifurcation ratios
    R_terminal          = (8*nu/np.pi)*terminal_length
    R_terminal_sister   = (8*nu/np.pi)*downstream_length + (data[edge,25]-(8*nu/np.pi)*data[edge,20])
    f_terminal          = (1+((data[edge,22]*R_terminal_sister)/(Qterm*R_terminal))**(gamma/4)) ** (-1/gamma)
    f_terminal_sister   = (1+((data[edge,22]*R_terminal_sister)/(Qterm*R_terminal))**(-gamma/4)) ** (-1/gamma)
    R_0                 = (8*nu/np.pi)*upstream_length + ((f_terminal**4)/R_terminal +
                          (f_terminal_sister**4)/R_terminal_sister)**-1
    L_0                 = f_terminal ** 2 * terminal_length + f_terminal_sister ** 2 * (downstream_length + data[edge,27])
    reduced_resistances.append(R_0)
    reduced_lengths.append(L_0)
    flows.append(data[edge,22].item()+Qterm)
    main_idx.append(edge.item())
    bifurcations.append(f_terminal_sister) # 0 -> 15; 1-> 16
    bifurcations.append(f_terminal)
    if edge == 0:
        # on the terminating round the new terminal is always 16 and the existing is 15
        #bifurcations.append([f_terminal_sister,f_terminal])
        L_0 = upstream_length + L_0
        R_0 = (((data[edge,22]+Qterm)*R_0)/(Pperm-Pterm))**(1/4)
        alt_idx.append(-1)
        new_scale.append(np.ones(f_terminal.shape))
        alt_scale.append(np.ones(f_terminal.shape))
        return (R_0,L_0,f_terminal,f_terminal_sister,None,mesh,reduced_resistances,
                reduced_lengths,bifurcations,flows,main_idx,alt_idx,new_scale,alt_scale,R_terminal,
                R_terminal_sister)
    previous = edge
    edge = int(data[edge,17].item())
    if int(data[edge,15].item()) == previous:
        alt = int(data[edge,16].item())
        if alt == -1:
            alt = -2
        alt_idx.append(alt)
        position = 0
        #bifurcations.append([f_terminal,f_terminal_sister])
    else:
        alt = int(data[edge,15].item())
        if alt == -1:
            alt = -2
        alt_idx.append(alt)
        position = 1
        #bifurcations.append([f_terminal_sister,f_terminal])
    if alt == -2:
        f_changed = np.array([1.0])
        f_stagnant = np.array([0.0])
        f_terminal = f_terminal*f_changed
        f_terminal_sister = f_terminal_sister*f_changed
        f_parent = f_changed
        R_0 = (8*nu/np.pi)*data[edge,20].item() + R_0
        L_0 = f_changed ** 2 *(upstream_length+L_0)
    else:
        f_changed = (1+((data[alt,22]*data[alt,25])/((data[previous,22]+Qterm)*R_0)) ** (gamma/4)) ** (-1/gamma)
        f_stagnant = (1+((data[alt,22]*data[alt,25])/((data[previous,22]+Qterm)*R_0)) ** (-gamma/4)) ** (-1/gamma)
        f_terminal = f_terminal*f_changed
        f_terminal_sister = f_terminal_sister*f_changed
        f_parent = f_changed
        R_0 = (8*nu/np.pi)*data[edge,20] + ((f_changed ** 4) /R_0 + (f_stagnant ** 4) /data[alt,25]) ** -1
        L_0 = f_changed ** 2 *(upstream_length+L_0) + f_stagnant ** 2 * (data[alt,20] + data[alt,27])
    reduced_resistances.append(R_0)
    reduced_lengths.append(L_0)
    flows.append(data[edge,22].item()+Qterm)
    main_idx.append(edge)
    new_scale.append(f_changed)
    alt_scale.append(f_stagnant)
    if position == 0:
        bifurcations.append(f_changed)
        bifurcations.append(f_stagnant)
    else:
        bifurcations.append(f_stagnant)
        bifurcations.append(f_changed)
    if edge == 0:
        L_0 = data[edge,20]+L_0
        R_0 = (((data[edge,22]+Qterm)*R_0)/(Pperm-Pterm))**(1/4)
        return (R_0,L_0,f_terminal,f_terminal_sister,f_parent,mesh,reduced_resistances,
                reduced_lengths,bifurcations,flows,main_idx,alt_idx,new_scale,alt_scale,R_terminal,
                R_terminal_sister)
    previous = edge
    edge = int(data[edge,17].item())
    while edge >= 0:
        if int(data[edge, 15].item()) == previous:
            alt = int(data[edge,16].item())
            if alt == -1:
                alt = -2
            alt_idx.append(alt)
            position = 0
        else:
            alt = int(data[edge,15].item())
            if alt == -1:
                alt = -2
            alt_idx.append(alt)
            position = 1
        if alt == -2:
            f_changed = np.array([1.0])
            f_stagnant = np.array([0.0])
            R_0 =  (8*nu/np.pi)*data[edge,20] + R_0
            L_0 =  f_changed ** 2 * (data[previous, 20] + L_0)
            f_terminal = f_terminal*f_changed
            f_terminal_sister = f_terminal_sister*f_changed
            f_parent = f_parent*f_changed
        else:
            f_changed = (1+((data[alt, 22]*data[alt,25])/((Qterm+data[previous,22])*R_0)) ** (gamma/4)) ** (-1/gamma)
            f_stagnant = (1+((data[alt, 22]*data[alt,25])/((Qterm+data[previous,22])*R_0)) ** (-gamma/4)) ** (-1/gamma)
            f_terminal = f_terminal*f_changed
            f_terminal_sister = f_terminal_sister*f_changed
            f_parent = f_parent*f_changed
            R_0 = (8*nu/np.pi)*data[edge,20] + ((f_changed ** 4 / R_0) + (f_stagnant ** 4 / data[alt, 25])) ** -1
            L_0 = f_changed ** 2 * (data[previous, 20] + L_0) + f_stagnant ** 2 *(data[alt, 20] + data[alt, 27])
        reduced_resistances.append(R_0)
        reduced_lengths.append(L_0)
        flows.append(data[edge,22].item()+Qterm)
        main_idx.append(edge)
        if edge >= 0:
            for j in range(len(new_scale)):
                new_scale[j] = new_scale[j]*f_changed
                alt_scale[j] = alt_scale[j]*f_changed
            new_scale.append(f_changed)
            #alt_scale = [bif*f_changed for bif in alt_scale]
            alt_scale.append(f_stagnant)
        #else:
        #    new_scale.append(np.ones(f_changed.shape))
        #    alt_scale.append(np.ones(f_changed.shape))
        if position == 0:
            bifurcations.append(f_changed)
            bifurcations.append(f_stagnant)
        else:
            bifurcations.append(f_stagnant)
            bifurcations.append(f_changed)
        previous = edge
        edge = int(data[edge,17].item())
    L_0 = data[0,20] + L_0
    R_0 = (((data[0, 22]+Qterm)*R_0)/(Pperm-Pterm))**(1/4)
    return (R_0,L_0,f_terminal,f_terminal_sister,f_parent,mesh,reduced_resistances,
            reduced_lengths,bifurcations,flows,main_idx,alt_idx,new_scale,alt_scale,R_terminal,
            R_terminal_sister)

"""
def fast_local_function_bind(bif_point,data,terminal,edge,gamma,nu,Qterm,Pperm,Pterm):
    upstream            = data[edge,0:3]
    downstream          = data[edge,3:6]
    upstream_length     = np.sqrt(np.sum(np.square(bif_point-upstream),axis=1))
    downstream_length   = np.sqrt(np.sum(np.square(bif_point-downstream),axis=1))
    terminal_length     = np.sqrt(np.sum(np.square(bif_point-terminal),axis=1))
"""
def path(data,edge):
    path_to_root = [edge]
    alt_paths    = []
    while path[-1] > 0:
        previous = edge
        edge     = int(data[edge,17].item())
        path_to_root.append(edge)
        left     = int(data[edge,15].item())
        right    = int(data[edge,16].item())
        if left == previous:
            alt_paths.append(right)
        else:
            alt_paths.append(left)
    return path_to_root,alt_paths

@nb.jit(nopython=True)
def volume(x,data,terminal,edge,gamma,nu,Qterm,Pperm,Pterm):
    ####################
    reduced_resistances = []
    reduced_lengths     = []
    bifurcations        = []
    flows               = [] # should only be changing path (not alt paths)
    main_idx            = [] # index 0: data[:,15]; index 1: data[:,16]
    alt_idx             = []
    new_scale           = [] # compound bifurcations
    alt_scale           = []
    rescale             = np.zeros(data.shape[0])
    ####################
    #total_flow          = data[edge,22].item()+Qterm*2
    upstream            = data[edge,0:3]
    downstream          = data[edge,3:6]
    upstream_length     = np.sqrt(np.sum(np.square(x-upstream),axis=1))
    downstream_length   = np.sqrt(np.sum(np.square(x-downstream),axis=1))
    terminal_length     = np.sqrt(np.sum(np.square(x-terminal),axis=1))
    # Begin calculating reduced resistance and bifurcation ratios
    R_terminal          = (8*nu/np.pi)*terminal_length
    R_terminal_sister   = (8*nu/np.pi)*downstream_length + (data[edge,25]-(8*nu/np.pi)*data[edge,20])
    f_terminal          = (1+((data[edge,22]*R_terminal_sister)/(Qterm*R_terminal))**(gamma/4)) ** (-1/gamma)
    f_terminal_sister   = (1+((data[edge,22]*R_terminal_sister)/(Qterm*R_terminal))**(-gamma/4)) ** (-1/gamma)
    R_0                 = (8*nu/np.pi)*upstream_length + ((f_terminal**4)/R_terminal +
                          (f_terminal_sister**4)/R_terminal_sister)**-1
    L_0                 = f_terminal ** 2 * terminal_length + f_terminal_sister ** 2 * (downstream_length + data[edge,27])
    reduced_resistances.append(R_0)
    reduced_lengths.append(L_0)
    flows.append(data[edge,22].item()+Qterm)
    main_idx.append(edge.item())
    bifurcations.append(f_terminal_sister) # 0 -> 15; 1-> 16
    bifurcations.append(f_terminal)
    if edge == 0:
        # on the terminating round the new terminal is always 16 and the existing is 15
        #bifurcations.append([f_terminal_sister,f_terminal])
        L_0 = upstream_length + L_0
        R_0 = (((data[edge,22]+Qterm)*R_0)/(Pperm-Pterm))**(1/4)
        alt_idx.append(-1)
        new_scale.append(np.ones(f_terminal.shape))
        alt_scale.append(np.ones(f_terminal.shape))
        return (R_0,L_0,f_terminal,f_terminal_sister,None,x,reduced_resistances,
                reduced_lengths,bifurcations,flows,main_idx,alt_idx,new_scale,alt_scale,R_terminal,
                R_terminal_sister)
    previous = edge
    edge = int(data[edge,17].item())
    if int(data[edge,15].item()) == previous:
        alt = int(data[edge,16].item())
        if alt == -1:
            alt = -2
        alt_idx.append(alt)
        position = 0
        #bifurcations.append([f_terminal,f_terminal_sister])
    else:
        alt = int(data[edge,15].item())
        if alt == -1:
            alt = -2
        alt_idx.append(alt)
        position = 1
        #bifurcations.append([f_terminal_sister,f_terminal])
    if alt == -2:
        f_changed = np.array([1.0])
        f_stagnant = np.array([0.0])
        f_terminal = f_terminal*f_changed
        f_terminal_sister = f_terminal_sister*f_changed
        f_parent = f_changed
        R_0 = (8*nu/np.pi)*data[edge,20].item() + R_0
        L_0 = f_changed ** 2 *(upstream_length+L_0)
    else:
        f_changed = (1+((data[alt,22]*data[alt,25])/((data[previous,22]+Qterm)*R_0)) ** (gamma/4)) ** (-1/gamma)
        f_stagnant = (1+((data[alt,22]*data[alt,25])/((data[previous,22]+Qterm)*R_0)) ** (-gamma/4)) ** (-1/gamma)
        f_terminal = f_terminal*f_changed
        f_terminal_sister = f_terminal_sister*f_changed
        f_parent = f_changed
        R_0 = (8*nu/np.pi)*data[edge,20] + ((f_changed ** 4) /R_0 + (f_stagnant ** 4) /data[alt,25]) ** -1
        L_0 = f_changed ** 2 *(upstream_length+L_0) + f_stagnant ** 2 * (data[alt,20] + data[alt,27])
    reduced_resistances.append(R_0)
    reduced_lengths.append(L_0)
    flows.append(data[edge,22].item()+Qterm)
    main_idx.append(edge)
    new_scale.append(f_changed)
    alt_scale.append(f_stagnant)
    if position == 0:
        bifurcations.append(f_changed)
        bifurcations.append(f_stagnant)
    else:
        bifurcations.append(f_stagnant)
        bifurcations.append(f_changed)
    if edge == 0:
        L_0 = data[edge,20]+L_0
        R_0 = (((data[edge,22]+Qterm)*R_0)/(Pperm-Pterm))**(1/4)
        return (R_0,L_0,f_terminal,f_terminal_sister,f_parent,x,reduced_resistances,
                reduced_lengths,bifurcations,flows,main_idx,alt_idx,new_scale,alt_scale,R_terminal,
                R_terminal_sister)
    previous = edge
    edge = int(data[edge,17].item())
    while edge >= 0:
        if int(data[edge, 15].item()) == previous:
            alt = int(data[edge,16].item())
            if alt == -1:
                alt = -2
            alt_idx.append(alt)
            position = 0
        else:
            alt = int(data[edge,15].item())
            if alt == -1:
                alt = -2
            alt_idx.append(alt)
            position = 1
        if alt == -2:
            f_changed = np.array([1.0])
            f_stagnant = np.array([0.0])
            R_0 =  (8*nu/np.pi)*data[edge,20] + R_0
            L_0 =  f_changed ** 2 * (data[previous, 20] + L_0)
            f_terminal = f_terminal*f_changed
            f_terminal_sister = f_terminal_sister*f_changed
            f_parent = f_parent*f_changed
        else:
            f_changed = (1+((data[alt, 22]*data[alt,25])/((Qterm+data[previous,22])*R_0)) ** (gamma/4)) ** (-1/gamma)
            f_stagnant = (1+((data[alt, 22]*data[alt,25])/((Qterm+data[previous,22])*R_0)) ** (-gamma/4)) ** (-1/gamma)
            f_terminal = f_terminal*f_changed
            f_terminal_sister = f_terminal_sister*f_changed
            f_parent = f_parent*f_changed
            R_0 = (8*nu/np.pi)*data[edge,20] + ((f_changed ** 4 / R_0) + (f_stagnant ** 4 / data[alt, 25])) ** -1
            L_0 = f_changed ** 2 * (data[previous, 20] + L_0) + f_stagnant ** 2 *(data[alt, 20] + data[alt, 27])
        reduced_resistances.append(R_0)
        reduced_lengths.append(L_0)
        flows.append(data[edge,22].item()+Qterm)
        main_idx.append(edge)
        if edge >= 0:
            for j in range(len(new_scale)):
                new_scale[j] = new_scale[j]*f_changed
                alt_scale[j] = alt_scale[j]*f_changed
            new_scale.append(f_changed)
            #alt_scale = [bif*f_changed for bif in alt_scale]
            alt_scale.append(f_stagnant)
        #else:
        #    new_scale.append(np.ones(f_changed.shape))
        #    alt_scale.append(np.ones(f_changed.shape))
        if position == 0:
            bifurcations.append(f_changed)
            bifurcations.append(f_stagnant)
        else:
            bifurcations.append(f_stagnant)
            bifurcations.append(f_changed)
        previous = edge
        edge = int(data[edge,17].item())
    L_0 = data[0,20] + L_0
    R_0 = (((data[0, 22]+Qterm)*R_0)/(Pperm-Pterm))**(1/4)
    return (R_0,L_0,f_terminal,f_terminal_sister,f_parent,x,reduced_resistances,
            reduced_lengths,bifurcations,flows,main_idx,alt_idx,new_scale,alt_scale,R_terminal,
            R_terminal_sister)


def constructor(data,terminal,edge,gamma,nu,Qterm,Pperm,Pterm,lam,mu,maxiter,method='L-BFGS-B'):
    proximal = data[edge,0:3]
    distal=data[edge,3:6]
    d_min    = min([np.linalg.norm(proximal-distal),
                    np.linalg.norm(terminal-distal),
                    np.linalg.norm(terminal-proximal)])*0.25
    d_sides    =   [np.linalg.norm(proximal-distal),
                    np.linalg.norm(terminal-distal),
                    np.linalg.norm(terminal-proximal)]
    s0 = (d_sides[0]+d_sides[1]+d_sides[2])/2
    area = (s0*abs(s0-d_sides[0])*abs(s0-d_sides[1])*abs(s0-d_sides[2]))**0.5
    #rescale  = deepcopy(data[:,28])
    def cost(st,func=volume,terminal=terminal,proximal=data[edge,0:3],distal=data[edge,3:6],data=data,edge=edge,gamma=gamma,nu=nu,Qterm=Qterm,Pperm=Pperm,Pterm=Pterm,d_min=d_min,d_sides=d_sides,area=area,lam=lam,mu=mu):
        s = st[0]
        t = st[1]
        if s > 1:
            s = 1
        elif s < 0:
            s = 0
        if t > 1:
            t = 1
        elif t < 0:
            t = 0
        x = proximal.reshape(1,-1)*(1-t)*s + \
            distal.reshape(1,-1)*(t*s)+\
            terminal.reshape(1,-1)*(1-s)
        d = [np.linalg.norm(proximal-x),
             np.linalg.norm(distal-x),
             np.linalg.norm(terminal-x)]
        dd = min(d)
        if dd < d_min:
            return 10000
        else:
            res = volume(x,data,terminal,edge,gamma,nu,Qterm,Pperm,Pterm)
            return np.pi*res[0]**lam*res[1]**mu
        """
        s1 = (d[0]+d[1]+d_sides[0])/2
        s2 = (d[1]+d[2]+d_sides[1])/2
        s3 = (d[2]+d[0]+d_sides[2])/2
        a1 = (s1*abs(s1-d[0])*abs(s1-d[1])*abs(s1-d_sides[0]))**0.5
        a2 = (s2*abs(s2-d[1])*abs(s2-d[2])*abs(s2-d_sides[1]))**0.5
        a3 = (s3*abs(s3-d[2])*abs(s3-d[0])*abs(s3-d_sides[2]))**0.5
        at = a1+a2+a3
        """
        #if not np.isclose(at,area):
        #    return 10000
        #else:
        #    res = volume(x,data,terminal,edge,gamma,nu,Qterm,Pperm,Pterm)
        #    return np.pi*res[0]**lam*res[1]**mu
    x0 = np.array([0.5,0.5])
    if method == 'Newton-CG' or method == 'dogleg' or method == 'trust-ncg':
        jac = lambda x: approx_fprime(x,cost,max(d_sides)/maxiter)
        res = minimize(cost,x0,method=method,options={'maxiter':maxiter},jac=jac)
    else:
        res = minimize(cost,x0,method=method,options={'maxiter':maxiter})
    st = res.x
    s = st[0]
    t = st[1]
    if s > 1:
        s = 1
    elif s < 0:
        s = 0
    if t > 1:
        t = 1
    elif t < 0:
        t = 0
    x = proximal.reshape(1,-1)*(1-t)*s + \
        distal.reshape(1,-1)*(t*s)+\
        terminal.reshape(1,-1)*(1-s)
    return volume(x,data,terminal,edge,gamma,nu,Qterm,Pperm,Pterm)



import numpy as np
import numba as nb
import vtk
import time
from .sphere_proximity import *
from copy import deepcopy
from .obb import *
"""
@nb.jit(nopython=True)
def separating_axis(position,plane,U1,V1,W1,U2,V2,W2,U1_scale,
                    V1_scale,W1_scale,U2_scale,V2_scale,W2_scale):
    return np.abs(np.dot(position,plane)) > (np.abs(np.dot(U1*U1_scale,plane))+
                                             np.abs(np.dot(V1*V1_scale,plane))+
                                             np.abs(np.dot(W1*W1_scale,plane))+
                                             np.abs(np.dot(U2*U2_scale,plane))+
                                             np.abs(np.dot(V2*V2_scale,plane))+
                                             np.abs(np.dot(W2*W2_scale,plane)))
"""
"""
#@nb.jit(nopython=True)
def obb2(data,edge):
    C1 = (edge[0:3] + edge[3:6]) / 2
    C2 = (data[:,0:3] + data[:,3:6]) / 2
    Position = C2 - C1

    # Basis vector construction
    edge[12:15] = ((edge[3:6] - edge[0:3]) /
                   np.linalg.norm(edge[3:6] -
                                  edge[0:3]))
    if edge[14] == -1:
        edge[6:9] = np.array([-1,0,0])
        edge[9:12] = np.array([0,-1,0])
    else:
        edge[6:9] = np.array([1-edge[12]**2/(1+edge[14]),
                              (-edge[12]*edge[13])/(1+edge[14]),
                              -edge[12]])
        edge[9:12] = np.array([(-edge[12]*edge[13])/(1+edge[14]),
                               1 - edge[13]**2/(1+edge[14]),
                               -edge[13]])
    result = []
    for i in range(data.shape[0]):
        # U1 axis
        if separating_axis(Position[i,:],edge[6:9],edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            continue
        # V1 axis
        if separating_axis(Position[i,:],edge[9:12],edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            continue
        # W1 axis
        if separating_axis(Position[i,:],edge[12:15],edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            continue
        # U2 axis
        if separating_axis(Position[i,:],data[i,6:9],edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            continue
        # V2 axis
        if separating_axis(Position[i,:],data[i,9:12],edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            continue
        # W2 axis
        if separating_axis(Position[i,:],data[i,12:15],edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            continue
        # Mixed Plane Separation Tests
        # U1 X U2
        plane = np.cross(edge[6:9],data[i,6:9])
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            continue
        # U1 X V2
        plane = np.cross(edge[6:9],data[i,9:12])
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            continue
        # U1 X W2
        plane = np.cross(edge[6:9],data[i,12:15])
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            continue
        # V1 X U2
        plane = np.cross(edge[9:12],data[i,6:9])
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            continue
        # V1 X V2
        plane = np.cross(edge[9:12],data[i,9:12])
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            continue
        # V1 X W2
        plane = np.cross(edge[9:12],data[i,12:15])
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            continue
        # W1 X U2
        plane = np.cross(edge[12:15],data[i,6:9])
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            continue
        # W1 X V2
        plane = np.cross(edge[12:15],data[i,9:12])
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            continue
        # W1 X W2
        plane = np.cross(edge[12:15],data[i,12:15])
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            continue
        else:
            pass
            #return True
    return False
"""

def collision_free(data,results,idx,terminal,vessel,radius_buffer):
    # proximal = 0
    # distal   = 1
    # terminal = 2
    new_vessels = np.ones((3,data.shape[1]))*-1
    new_vessels[0,0:3] = data[vessel,0:3]
    new_vessels[0,3:6] = results[5][idx]
    new_vessels[0,-1] = data[vessel,-1]
    if vessel == 0:
        new_vessels[0,21]  = results[0][idx] + radius_buffer
    else:
        new_vessels[0,21]  = results[0][idx]*results[4][idx] + radius_buffer
        new_vessels[0,17]  = data[vessel,17]
    new_vessels[1,0:3] = results[5][idx]
    new_vessels[1,3:6] = data[vessel,3:6]
    new_vessels[1,21]  = results[0][idx]*results[3][idx] + radius_buffer
    new_vessels[1,15:17] = data[vessel,15:17]
    new_vessels[1,17]  = np.float(vessel)
    new_vessels[2,0:3] = results[5][idx]
    new_vessels[2,3:6] = terminal
    new_vessels[2,21]  = results[0][idx]*results[2][idx] + radius_buffer
    new_vessels[2,17]  = np.float(vessel)
    #print(new_vessels[:,-1])
    vessel_parent = data[vessel,17]
    vessel_child1 = data[vessel,15]
    vessel_child2 = data[vessel,16]
    vessel_sister_i = data[int(vessel_parent),15]
    vessel_sister_j = data[int(vessel_parent),16]
    for i in range(new_vessels.shape[0]):
        ##Check Self
        #start = time.time()
        n = sphere_proximity(data,new_vessels[i,:])
        #print(time.time()-start)
        if len(np.argwhere(n==vessel)) > 0:
            n = np.delete(n,np.argwhere(n==vessel).flatten())
        if len(np.argwhere(n==int(vessel_parent))) > 0:
            n = np.delete(n,np.argwhere(n==int(vessel_parent)).flatten())
        if len(np.argwhere(n==int(vessel_child1))) > 0:
            n = np.delete(n,np.argwhere(n==int(vessel_child1)).flatten())
        if len(np.argwhere(n==int(vessel_child2))) > 0:
            n = np.delete(n,np.argwhere(n==int(vessel_child2)).flatten())
        if len(np.argwhere(n==int(vessel_sister_i))) > 0:
            n = np.delete(n,np.argwhere(n==int(vessel_sister_i)).flatten())
        if len(np.argwhere(n==int(vessel_sister_j))) > 0:
            n = np.delete(n,np.argwhere(n==int(vessel_sister_j)).flatten())
        if len(np.argwhere(n==int(new_vessels[i,-1]))) > 0:
            n = np.delete(n,np.argwhere(n==int(new_vessels[i,-1])).flatten())
            #print('self')
        #Check Parent
        if len(np.argwhere(n==int(new_vessels[i,17]))) > 0:
            n = np.delete(n,np.argwhere(n==int(new_vessels[i,17])).flatten())
            #print('parent')
        #Check Children
        if len(np.argwhere(n==int(new_vessels[i,15]))) > 0:
            n = np.delete(n,np.argwhere(n==int(new_vessels[i,15])).flatten())
            #print('children')
        if len(np.argwhere(n==int(new_vessels[i,16]))) > 0:
            n = np.delete(n,np.argwhere(n==int(new_vessels[i,16])).flatten())
            #print('children')
        #Check Sister
        if int(new_vessels[i,17]) != -1:
            parent = int(new_vessels[i,17])
            if len(np.argwhere(n==int(data[parent,15]))) > 0:
                n = np.delete(n,np.argwhere(n==int(data[parent,15])).flatten())
            if len(np.argwhere(n==int(data[parent,16]))) > 0:
                n = np.delete(n,np.argwhere(n==int(data[parent,16])).flatten())
                #print('sister')
        #else:
        #    if len(np.argwhere(n==int(new_vessels[i,-1]))) > 0:
        #        n = np.delete(n,np.argwhere(n==int(new_vessels[i,-1])).flatten())

        if len(n) == 0:
            continue
        else:
            if obb(data[n,:],new_vessels[i,:]):
                #print(n)
                #print(new_vessels[:,-1])
                """
                colors = vtk.vtkNamedColors()
                models = []
                actors = []
                background_color='white'
                resolution=100
                center = tuple((new_vessels[i,0:3] + new_vessels[i,3:6])/2)
                radius = new_vessels[i,21]
                direction = tuple(new_vessels[i,12:15])
                vessel_length = new_vessels[i,20]
                cyl = vtk.vtkTubeFilter()
                line = vtk.vtkLineSource()
                line.SetPoint1(new_vessels[i,0],new_vessels[i,1],new_vessels[i,2])
                line.SetPoint2(new_vessels[i,3],new_vessels[i,4],new_vessels[i,5])
                cyl.SetInputConnection(line.GetOutputPort())
                cyl.SetRadius(radius)
                cyl.SetNumberOfSides(resolution)
                models.append(cyl)
                mapper = vtk.vtkPolyDataMapper()
                actor  = vtk.vtkActor()
                mapper.SetInputConnection(cyl.GetOutputPort())
                actor.SetMapper(mapper)
                actor.GetProperty().SetColor(colors.GetColor3d('green'))
                actors.append(actor)
                for edge in n:
                    center = tuple((data[edge,0:3] + data[edge,3:6])/2)
                    radius = data[edge,21]
                    direction = tuple(data[edge,12:15])
                    vessel_length = data[edge,20]
                    cyl = vtk.vtkTubeFilter()
                    line = vtk.vtkLineSource()
                    line.SetPoint1(data[edge,0],data[edge,1],data[edge,2])
                    line.SetPoint2(data[edge,3],data[edge,4],data[edge,5])
                    cyl.SetInputConnection(line.GetOutputPort())
                    cyl.SetRadius(radius)
                    cyl.SetNumberOfSides(resolution)
                    models.append(cyl)
                    mapper = vtk.vtkPolyDataMapper()
                    actor  = vtk.vtkActor()
                    mapper.SetInputConnection(cyl.GetOutputPort())
                    actor.SetMapper(mapper)
                    if edge in n:
                        #print('True')
                        actor.GetProperty().SetColor(colors.GetColor3d('blue'))
                    if edge == int(new_vessels[i,-1]):
                        print(edge)
                        actor.GetProperty().SetColor(colors.GetColor3d('green'))
                    else:
                        actor.GetProperty().SetColor(colors.GetColor3d('red'))
                    actors.append(actor)
                renderer = vtk.vtkRenderer()
                renderer.SetBackground(colors.GetColor3d(background_color))

                render_window = vtk.vtkRenderWindow()
                render_window.AddRenderer(renderer)
                render_window.SetWindowName('SimVascular Vessel Collision')

                interactor = vtk.vtkRenderWindowInteractor()
                interactor.SetRenderWindow(render_window)
                for actor in actors:
                     renderer.AddActor(actor)

                render_window.Render()
                interactor.Start()
                """
                return False
    return True

def no_outside_collision(tree,outside_vessels,radius_buffer=0):
    outside_vessels[:,21] += radius_buffer
    for i in range(outside_vessels.shape[0]):
        n = sphere_proximity(tree.data,outside_vessels[i,:])
        if len(n) == 0:
            continue
        else:
            if obb(tree.data[n,:],outside_vessels[i,:]):
                return False
    return True

def collision_free_fd(data,vessel):
    ii       = np.array(list(range(data.shape[0]-2)))
    ii       = ii[ii!=vessel]
    tree     = data[:-2,:]
    new      = data[-2:,:]
    proximal = data[vessel,:]
    new_vessels = np.vstack((new,proximal))
    for i in range(new_vessels.shape[0]):
        ##Check Self
        n = sphere_proximity(tree,new_vessels[i,:])
        if len(np.argwhere(n==int(new_vessels[i,-1]))) > 0:
            n = np.delete(n,np.argwhere(n==int(new_vessels[i,-1])).flatten())
        #Check Parent
        if len(np.argwhere(n==int(new_vessels[i,17]))) > 0:
            n = np.delete(n,np.argwhere(n==int(new_vessels[i,17])).flatten())
        #Check Children
        if len(np.argwhere(n==int(new_vessels[i,15]))) > 0:
            n = np.delete(n,np.argwhere(n==int(new_vessels[i,15])).flatten())
        if len(np.argwhere(n==int(new_vessels[i,16]))) > 0:
            n = np.delete(n,np.argwhere(n==int(new_vessels[i,16])).flatten())
        #Check Sister
        if int(new_vessels[i,17]) != -1:
            parent = int(new_vessels[i,17])
            if len(np.argwhere(n==int(data[parent,15]))) > 0:
                n = np.delete(n,np.argwhere(n==int(data[parent,15])).flatten())
            if len(np.argwhere(n==int(data[parent,16]))) > 0:
                n = np.delete(n,np.argwhere(n==int(data[parent,16])).flatten())
        #else:
        #    if len(np.argwhere(n==int(new_vessels[i,-1]))) > 0:
        #        n = np.delete(n,np.argwhere(n==int(new_vessels[i,-1])).flatten())

        if len(n) == 0:
            continue
        else:
            if obb(tree[n,:],new_vessels[i,:]):
                """
                colors = vtk.vtkNamedColors()
                models = []
                actors = []
                background_color='white'
                resolution=100
                for edge in range(new_vessels.shape[0]):
                    center = tuple((new_vessels[edge,0:3] + new_vessels[edge,3:6])/2)
                    radius = new_vessels[edge,21]
                    direction = tuple(new_vessels[edge,12:15])
                    vessel_length = new_vessels[edge,20]
                    cyl = vtk.vtkTubeFilter()
                    line = vtk.vtkLineSource()
                    line.SetPoint1(new_vessels[edge,0],new_vessels[edge,1],new_vessels[edge,2])
                    line.SetPoint2(new_vessels[edge,3],new_vessels[edge,4],new_vessels[edge,5])
                    cyl.SetInputConnection(line.GetOutputPort())
                    cyl.SetRadius(radius)
                    cyl.SetNumberOfSides(resolution)
                    models.append(cyl)
                    mapper = vtk.vtkPolyDataMapper()
                    actor  = vtk.vtkActor()
                    mapper.SetInputConnection(cyl.GetOutputPort())
                    actor.SetMapper(mapper)
                    actor.GetProperty().SetColor(colors.GetColor3d('red'))
                    actors.apend(actor)
                for edge in n:
                    center = tuple((data[edge,0:3] + data[edge,3:6])/2)
                    radius = data[edge,21]
                    direction = tuple(data[edge,12:15])
                    vessel_length = data[edge,20]
                    cyl = vtk.vtkTubeFilter()
                    line = vtk.vtkLineSource()
                    line.SetPoint1(data[edge,0],data[edge,1],data[edge,2])
                    line.SetPoint2(data[edge,3],data[edge,4],data[edge,5])
                    cyl.SetInputConnection(line.GetOutputPort())
                    cyl.SetRadius(radius)
                    cyl.SetNumberOfSides(resolution)
                    models.append(cyl)
                    mapper = vtk.vtkPolyDataMapper()
                    actor  = vtk.vtkActor()
                    mapper.SetInputConnection(cyl.GetOutputPort())
                    actor.SetMapper(mapper)
                    if edge in n:
                        actor.GetProperty().SetColor(colors.GetColor3d('blue'))
                    if edge == int(new_vessels[i,-1]):
                        actor.GetProperty().SetColor(colors.GetColor3d('green'))
                    else:
                        actor.GetProperty().SetColor(colors.GetColor3d('red'))
                    actors.append(actor)
                renderer = vtk.vtkRenderer()
                renderer.SetBackground(colors.GetColor3d(background_color))

                render_window = vtk.vtkRenderWindow()
                render_window.AddRenderer(renderer)
                render_window.SetWindowName('SimVascular Vessel Collision')

                interactor = vtk.vtkRenderWindowInteractor()
                interactor.SetRenderWindow(render_window)
                for actor in actors:
                     renderer.AddActor(actor)

                render_window.Render()
                interactor.Start()
                """
                return False
    return True


def line_distance(data,edge):
    a = np.sum((data[:,3:6]-data[:,0:3])*(data[:,3:6]-data[:,0:3]),axis=1)
    b = np.dot(data[:,3:6]-data[:,0:3],edge[3:6].T-edge[0:3].T)
    c = np.dot(edge[3:6]-edge[0:3],edge[3:6].T-edge[0:3].T)
    d = np.sum((data[:,3:6]-data[:,0:3])*(data[:,0:3]-edge[0:3]),axis=1)
    e = np.dot(edge[3:6]-edge[0:3],(data[:,0:3]-edge[0:3]).T)
    f00 = d
    f10 = f00 + a
    f01 = f00 + b
    f11 = f10 - b

    g00 = -e
    g10 = g00 - b
    g01 = g00 + c
    g11 = g10 + c

    p0 = np.zeros((len(f00),3))
    p1 = np.zeros((len(f00),3))
    dist = np.zeros((len(f00),1))
    for j in range(len(f00)):
        if a[j] > 0 and c > 0:
            s = [0,0]
            s[0] = clamped_root(a[j],f00[j],f10[j])
            s[1] = clamped_root(a[j],f01[j],f11[j])
            idd = [0,0]
            parameters = [0,0]
            for ii in range(2):
                if s[ii] <= 0:
                    idd[ii] = -1
                elif s[ii] >= 1:
                    idd[ii] = 1
                else:
                    idd[ii] = 0
            if idd[0] == -1 and idd[1] == -1:
                parameters[0] = 0
                parameters[1] = clamped_root(c,g00[j],g01[j])
            elif idd[0] == 1 and idd[0] == 1:
                parameters[0] = 1
                parameters[1] = clamped_root(c,g10[j],g11[j])
            else:
                ed,M = line_intersection(s,idd,f10=f10[j],f00=f00[j],b=b[j])
                parameters = find_minimum(ed,M,b=b[j],c=c,e=e[j],g00=g00[j],
                                          g01=g01[j],g10=g10[j],g11=g11[j])
        else:
            if a[j] > 0:
                parameters[0] = clamped_root(a[j],f00[j],f10[j])
                parameters[1] = 0
            elif c > 0:
                parameters[0] = 0
                parameters[1] = clamped_root(c,g00[j],g01[j])
            else:
                parameters[0] = 0
                parameters[0] = 0
        p0[j,:] = (1 - parameters[0])*data[j,0:3] + parameters[0]*data[j,3:6]
        p1[j,:] = (1 - parameters[1])*edge[0:3] + parameters[1]*edge[3:6]
        dist[j] = np.sqrt(np.sum(np.square(p0[j,:]-p1[j,:])))
    return p0,p1,dist

def clamped_root(slope,h0,h1):
    if h0 < 0:
        if h1 > 0:
            r = -h0/slope
            if r > 1:
                r = 0.5
        else:
            r = 1
    else:
        r = 0
    return r

def line_intersection(s,id,f10=None,f00=None,b=None):
    edge = [0,0]
    M = np.zeros((2,2))
    if id[0] < 0:
        edge[0] = 0
        M[0,0] = 0
        M[0,1] = f00/b
        if M[0,1] < 0 or M[0,1] > 1:
            M[0,1] = 0.5
        if id[1] == 0:
            edge[1] = 3
            M[1,0] = s[1]
            M[1,1] = 1
        else:
            edge[1] = 1
            M[1,0] = 1
            M[1,1] = f10/b
            if M[1,1] < 0 or M[1,1] > 1:
                M[1,1] = 0.5
    elif id[0] == 0:
        edge[0] = 2
        M[0,0] = s[0]
        M[0,1] = 0

        if id[1] < 0:
            edge[1] = 0
            M[1,0] = 0
            M[1,1] = f00/b
            if M[1,1] < 0 or M[1,1] > 1:
                M[1,1] = 0.5
        elif id[1] == 0:
            edge[1] = 3
            M[1,0] = s[1]
            M[1,1] = 1
        else:
            edge[1] = 1
            M[1,0] = 1
            M[1,1] = f10/b
            if M[1,1] < 0 or M[1,1] > 1:
                M[1,1] = 0.5
    else:
        edge[0] = 1
        M[0,0] = 1
        M[0,1] = f10/b
        if M[0,1] < 0 or M[0,1] > 1:
            M[0,1] = 0.5
        if id[1] == 0:
            edge[1] = 3
            M[1,0] = s[1]
            M[1,1] = 1
        else:
            edge[1] = 0
            M[1,0] = 0
            M[1,1] = f00/b
            if M[1,1] < 0 or M[1,1] > 1:
                M[1,1] = 0.5
    return edge, M


def find_minimum(edge,M,b=None,c=None,e=None,
                 g00=None,g01=None,g10=None,g11=None):
    delta = M[1,1] - M[0,1]
    h0 = delta * (-b * M[0,0] + c*M[0,1] - e)
    parameters = [0,0]
    if h0 >= 0:
        if edge[0] == 0:
            parameters[0] = 0
            parameters[1] = clamped_root(c,g00,g01)
        elif edge[0] == 1:
            parameters[0] = 1
            parameters[1] = clamped_root(c,g10,g11)
        else:
            parameters[0] = M[0,0]
            parameters[1] = M[0,1]
    else:
        h1 = delta*(-b * M[1,0] + c*M[1,1] - e)
        if h1 <= 0:
            if edge[1] == 0:
                parameters[0] = 0
                parameters[1] = clamped_root(c,g00,g01)
            elif edge[1] == 1:
                parameters[0] = 1
                parameters[1] = clamped_root(c,g10,g11)
            else:
                parameters[0] = M[1,0]
                parameters[1] = M[1,1]
        else:
            k = min(max(h0/(h0-h1),0),1)
            mk = 1 - k
            parameters[0] = mk*M[0,0]+k*M[1,0]
            parameters[1] = mk*M[0,1]+k*M[1,1]
    return parameters

def pairwise_tree_collisions(tree1_data,tree2_data,radius_buffer=0):
    outside_vessels = deepcopy(tree2_data)
    outside_vessels[:,21] += radius_buffer
    pairs = []
    for i in range(outside_vessels.shape[0]):
        n = sphere_proximity(tree1_data,outside_vessels[i,:])
        if len(n) == 0:
            continue
        else:
            results = obbc(tree1_data[n,:],outside_vessels[i,:])
            if len(results) == 0:
                continue
            else:
                for r in results:
                    pairs.append([r,i])
    return pairs

def fix_suggestions(tree1_data,tree2_data,radius_buffer=0):
    pairs = pairwise_tree_collisions(tree1_data,tree2_data,radius_buffer=radius_buffer)
    pair_points     = []
    pair_lengths    = []
    pair_adjustment = []
    pair_vectors    = []
    for pr in pairs:
        p0,p1,length = line_distance(tree1_data[pr[0],:].reshape(1,-1),tree2_data[pr[1],:])
        pair_points.append([p0,p1])
        pair_lengths.append(length)
        pair_adjustment.append([abs(tree1_data[pr[0],21]+tree2_data[pr[1],21]+radius_buffer)/2,
                                abs(tree1_data[pr[0],21]+tree2_data[pr[1],21]+radius_buffer)/2])
        pair_vectors.append([(p0-p1)/np.linalg.norm(p0-p1),(p1-p0)/np.linalg.norm(p1-p0)])
    return pairs,pair_points,pair_lengths,pair_adjustment,pair_vectors



# Mesh searching versus kd-VVU patch searching
import svcco
import pyvista as pv
from time import perf_counter
import os
import numpy as np
import matplotlib.pyplot as plt
from tqdm import trange
from scipy.optimize import curve_fit
###########################################
# Code for Building Surface
###########################################

cube = pv.Cube().triangulate().subdivide(3)

s = svcco.surface()
s.set_data(10*cube.points,cube.point_normals)
s.solve()
s.build()

###########################################
#
###########################################

def test(surf,point_range=(1,100),iter=25):
    mesh = surf.tet.grid
    mesh_search = surf.cell_lookup
    kd   = surf.patch_KDTree
    times_mesh = []
    times_kd   = []
    xb = mesh.bounds[1] - mesh.bounds[0]
    yb = mesh.bounds[3] - mesh.bounds[2]
    zb = mesh.bounds[5] - mesh.bounds[4]
    mesh_search.query([0,0,0])
    kd.query([0,0,0])
    for i in range(iter):
        timeline_mesh = []
        timeline_kd   = []
        for j in trange(point_range[0],point_range[1]):
            total_time_mesh = 0
            total_time_kd   = 0
            #for k in range(j):
            xp = (mesh.bounds[0] + np.random.random(j)*xb)
            yp = (mesh.bounds[2] + np.random.random(j)*yb)
            zp = (mesh.bounds[4] + np.random.random(j)*zb)
            point = np.array([xp,yp,zp]).T
            start = perf_counter()
            mesh_search.query(point)
            total_time_mesh += perf_counter()-start
            start = perf_counter()
            kd.query(point)
            total_time_kd += perf_counter()-start
            timeline_mesh.append(total_time_mesh)
            timeline_kd.append(total_time_kd)
        times_mesh.append(timeline_mesh)
        times_kd.append(timeline_kd)
    times_mesh = np.array(times_mesh)
    times_kd   = np.array(times_kd)
    time_mesh_avg = np.mean(times_mesh,axis=0)
    time_mesh_std = np.std(times_mesh,axis=0)
    time_kd_avg = np.mean(times_kd,axis=0)
    time_kd_std = np.std(times_kd,axis=0)
    return time_mesh_avg,time_mesh_std,time_kd_avg,time_kd_std
#test(s)

def gen_figure(surf_object,point_range=(1,100),iter=25):
    fig = plt.figure()
    ax  = fig.add_subplot(111)
    TM,TM_STD,TP,TP_STD = test(surf_object,point_range=point_range,iter=iter)
    PTS = list(range(point_range[0],point_range[1]))
    ax.plot(PTS,TM,label="Mesh",color='red')
    ax.fill_between(PTS,TM-2*TM_STD,TM+TM_STD,alpha=0.5,color='red')
    ax.plot(PTS,TP,label="VVU",color='blue')
    ax.fill_between(PTS,TP-TP_STD,TP+TP_STD,alpha=0.5,color='blue')
    ax.set_xlabel("Number of Points")
    ax.set_ylabel("Time (seconds)")
    #ax.set_ylim(bottom=0)
    ax.set_xlim(left=0,right=point_range[1])
    ax.set_yscale('log')
    plt.legend()
    name = "mesh_vs_patches_point-range_{}-{}_iter_{}.svg".format(point_range[0],point_range[1],iter)
    fig.savefig(name,format="svg")
    plt.show()
    return TM,TM_STD,TP,TP_STD

def gen_figure_curve_fit(surf_object,point_range=(1,100),iter=25,func=None):
    fig = plt.figure()
    ax  = fig.add_subplot(111)
    TM,TM_STD,TP,TP_STD = test(surf_object,point_range=point_range,iter=iter)
    if func is None:
        def func(x,a,b,c,d,e):
            return  a*(np.log(b*x+c)**d)+e
    PTS = list(range(point_range[0],point_range[1]))
    ax.plot(PTS,TM,label="Mesh",color='red',linestyle='dashed')
    ax.fill_between(PTS,TM-2*TM_STD,TM+TM_STD,alpha=0.5,color='red')
    ax.plot(PTS,TP,label="VVU",color='blue',linestyle='dashed')
    ax.fill_between(PTS,TP-TP_STD,TP+TP_STD,alpha=0.5,color='blue')
    ax.set_xlabel("Number of Points")
    ax.set_ylabel("Time (seconds)")
    #ax.set_ylim(bottom=0)
    ax.set_xlim(left=0,right=point_range[1])
    ax.set_yscale('log')
    # curve fitting
    PTS = np.array(PTS)
    popt_mesh,pcov_mesh = curve_fit(func,PTS,TM,p0=[1,1,1,1,1],maxfev=50000)
    popt_pat,pcov_pat = curve_fit(func,PTS,TP,p0=[1,1,1,1,1],maxfev=50000)
    ax.plot(PTS,func(PTS,popt_mesh[0],popt_mesh[1],popt_mesh[2],popt_mesh[3],popt_mesh[4]),color='red',linestyle='solid')
    ax.plot(PTS,func(PTS,popt_pat[0],popt_pat[1],popt_pat[2],popt_pat[3],popt_pat[4]),color='blue',linestyle='solid')
    plt.legend()
    name = "mesh_vs_patches_point-range_{}-{}_iter_{}_with_curve.svg".format(point_range[0],point_range[1],iter)
    fig.savefig(name,format="svg")
    plt.show()
    return TM,TM_STD,TP,TP_STD,popt_mesh,pcov_mesh,popt_pat,pcov_pat
"""
def results(start=3,end=6):
    POINTS = []
    MESH_AVG   = []
    MESH_STD   = []
    KD_AVG     = []
    KD_STD     = []
    for i in range(start,end):
        print("{}/{}".format(i,end-1))
        cube = pv.Cube().triangulate().subdivide(i)
        s = svcco.surface()
        s.set_data(10*cube.points,normals=cube.point_normals)
        s.solve()
        s.build()
        POINTS.append(s.points.shape[0])
        m_avg,m_std,kd_avg,kd_std = test(s)
        MESH_AVG.append(m_avg)
        MESH_STD.append(m_std)
        KD_AVG.append(kd_avg)
        KD_STD.append(kd_std)
    MESH_AVG = np.array(MESH_AVG)
    MESH_STD = np.array(MESH_STD)
    KD_AVG   = np.array(KD_AVG)
    KD_STD   = np.array(KD_STD)
    fig = plt.figure()
    ax  = fig.add_subplot(111)
    ax.plot(POINTS,MESH_AVG,label="Mesh")
    ax.fill_between(POINTS,MESH_AVG-MESH_STD,MESH_AVG+MESH_STD,alpha=0.5)
    ax.plot(POINTS,KD_AVG,label="kd-VVU")
    ax.fill_between(POINTS,KD_AVG-KD_STD,KD_AVG+KD_STD,alpha=0.5)
    ax.set_xlabel("Number of Points")
    ax.set_ylabel("Time (seconds)")
    ax.set_ylim(bottom=0)
    plt.legend()
    plt.show()
"""



import numpy as np
import svcco
from time import perf_counter as time
import matplotlib.pyplot as plt
from tqdm import tqdm

def global_pu_time(start,stop):
    sizes = np.arange(start,stop)
    num_points = []
    pu_time = []
    gu_time = []
    pur_time = []
    gur_time = []
    #warmup the code for jit compilation
    sph = svcco.implicit.tests.bumpy_sphere.bumpy_sphere(10)
    s = svcco.surface()
    s.set_data(sph)
    s.solve()
    s.solve(PU=False)
    for si in tqdm(sizes):
        sph = svcco.implicit.tests.bumpy_sphere.bumpy_sphere(samples=si)
        num_points.append(sph.shape[0])
        s = svcco.surface()
        s.set_data(sph)
        begin = time()
        s.solve()
        pu_time.append(time() - begin)
        #begin = time()
        #s.solve(regularize=True)
        #pur_time.append(time() - begin)
        begin = time()
        s.solve(PU=False)
        gu_time.append(time() - begin)
        #begin = time()
        #s.solve(PU=False,regularize=True)
        #gur_time.append(time() - begin)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.loglog(num_points,pu_time,label='Partition of Unity')
    #ax.plot(num_points,pur_time,label='Regularized Partition of Unity')
    ax.loglog(num_points,gu_time,label='Global')
    ax.set_xlabel('Number of Points')
    ax.set_ylabel('Time (sec)')
    #ax.plot(num_points,gur_time,label='Regularize Global')
    plt.grid(True, which="both", ls="-")
    plt.legend()
    plt.savefig('PU_vs_Global_Time_plot.svg', format="svg", bbox_inches='tight')
    return num_points,pu_time,pur_time,gu_time,gur_time,



# Script to obtain build-time data for convex cone geometry
import pyvista as pv
import os
from pickle import dump
import svcco
import numpy as np

name = 'octahedron'
shape = pv.Octahedron().triangulate().subdivide(3)

######################
number = 100
######################

FOLDER = 'convex_data'

if not os.path.isdir(os.getcwd()+os.sep+FOLDER):
    os.mkdir(os.getcwd()+os.sep+FOLDER)

if not os.path.isdir(os.getcwd()+os.sep+FOLDER+os.sep+name):
    os.mkdir(os.getcwd()+os.sep+FOLDER+os.sep+name)
    os.mkdir(os.getcwd()+os.sep+FOLDER+os.sep+name+os.sep+'times')
    os.mkdir(os.getcwd()+os.sep+FOLDER+os.sep+name+os.sep+'data')

s = svcco.surface()
s.set_data(10*shape.points,normals=shape.point_normals)
s.solve()
s.build()


# Warm-up build to catch jit-compilation of functions
t = svcco.tree()
t.set_boundary(s)
t.parameters['Qterm'] *= 40
t.set_assumptions(convex=True)
t.set_root()
t.n_add(400)

EXISTING_TIMES     = len(os.listdir(os.getcwd()+os.sep+FOLDER+os.sep+name+os.sep+'times'))
EXISTING_DATA      = len(os.listdir(os.getcwd()+os.sep+FOLDER+os.sep+name+os.sep+'data'))
EXISTING_PERFUSION = len(list(filter(lambda n: 'perfusion' in n,os.listdir(os.getcwd()+os.sep+FOLDER+os.sep+name+os.sep+'data'))))
EXISTING_MESH      = len(list(filter(lambda n: 'mesh' in n,os.listdir(os.getcwd()+os.sep+FOLDER+os.sep+name+os.sep+'data'))))
EXISTING_TREE      = len(list(filter(lambda n: 'tree' in n,os.listdir(os.getcwd()+os.sep+FOLDER+os.sep+name+os.sep+'data'))))
EXISTING_RAW       = len(list(filter(lambda n: 'raw' in n,os.listdir(os.getcwd()+os.sep+FOLDER+os.sep+name+os.sep+'data'))))

for i in range(number):
    t = svcco.tree()
    t.set_boundary(s)
    t.parameters['Qterm'] *= 40
    t.set_assumptions(convex=True)
    t.set_root()
    t.n_add(8000)
    file = open(os.getcwd() + os.sep +FOLDER +os.sep+name+os.sep+'times'+os.sep+'{}_times_8000_vessels_{}.pkl'.format(name,i+EXISTING_TIMES),"wb")
    dump(t.time,file)
    file.close()

    id,vol = svcco.perfusion_territory(t,mesh_file=os.getcwd()+os.sep+FOLDER+os.sep+name+os.sep+'data'+os.sep+name+'_mesh_{}.vtu'.format(i+EXISTING_MESH),
                                         tree_file=os.getcwd()+os.sep+FOLDER+os.sep+name+os.sep+'data'+os.sep+name+'_tree_{}.vtp'.format(i+EXISTING_TREE))
    file = open(os.getcwd() + os.sep + FOLDER +os.sep+name+os.sep+'data'+os.sep+name+'_perfusion_{}.pkl'.format(i+EXISTING_PERFUSION),"wb")
    dump(vol,file)
    file.close()

    np.savetxt(os.getcwd() + os.sep + FOLDER +os.sep+name+os.sep+'data'+os.sep+name+'_raw_{}.csv'.format(i+EXISTING_RAW),t.data,delimiter=',')

grid = s.pv_polydata.delaunay_3d()
surf = grid.extract_surface()
stats = [s.pv_polydata.area,s.pv_polydata.volume,surf.area,grid.volume]
file = open(os.getcwd() + os.sep + FOLDER+os.sep+name+os.sep+'data'+os.sep+'stats.pkl',"wb")
dump(stats,file)
file.close()



import numpy as np
from pickle import load
import matplotlib.pyplot as plt
from matplotlib import cm
from tqdm import tqdm
import os

###############################################
# Parameters to change for figure display
###############################################

tree_sizes = [100,1000,4000,7999]

#convex_directories    = [os.getcwd()+os.sep+'cube_times',os.getcwd()+os.sep+'sphere_times']
convex_directories = os.walk(os.getcwd()+os.sep+'convex_data')
convex_directories = [x[0] for x in convex_directories]
convex_directories = list(filter(lambda n: 'times' in n,convex_directories))

nonconvex_directories = os.walk(os.getcwd()+os.sep+'brain_data')
nonconvex_directories = [x[0] for x in nonconvex_directories]
nonconvex_directories = list(filter(lambda n: 'data' in n,nonconvex_directories))

#nonconvex_directories = os.walk(os.getcwd()+os.sep+'brain_data')
#nonconvex_directories = [x[0] for x in nonconvex_directories]
#nonconvex_directories = list(filter(lambda n: 'times' in n,nonconvex_directories))

#linecolors = []
#red = cm.Reds
#for l in range(len(convex_directories)):
#    linecolors.append(red(256-l))

#blue = cm.Blues
#for l in range(len(nonconvex_directories)):
#    linecolors.append(blue(256-l))

#convex_directories.extend(nonconvex_directories)
#directories  = convex_directories

directories = nonconvex_directories

#linestyles = [(0,())]*len(directories)

#for l in range(len(nonconvex_directories)):
#    if l == 0:
#        linestyles.append(tuple([0,tuple([])]))
#    else:
#        linestyles.append(tuple([0,tuple([3,l])]))

#capsize = 5

###############################################
# Below is just code to unpack data
# there is no need to touch this.
###############################################

DATA = []
data = []

for dirname in directories:
    if 'stats.pkl' not in os.listdir(dirname):
        continue
    file = dirname + os.sep + 'stats.pkl'
    f = open(file,'rb')
    conn = load(f)
    DATA.append((conn[3])/(conn[1]))
    #if DATA[-1] > 2:
    #    print(dirname)
    dir_adjust = dirname.split(os.sep)[:-1]
    name = ''
    for n in dir_adjust:
        name += n + os.sep
    dir_adjust = name + 'times'
    files = os.listdir(dir_adjust)
    #print(dir_adjust)
    tmp = np.zeros((len(files),len(tree_sizes)))
    for i, fname in tqdm(enumerate(files),desc='loading {} directory'.format(dir_adjust)):
        f = open(dir_adjust+os.sep+fname,'rb')
        time_dat = load(f)
        for j, size in enumerate(tree_sizes):
            tmp[i,j] = sum(time_dat['total'][:size])
    data.append(np.mean(tmp[:,-1]))
    #data[dirname+'_std']  = np.std(data[dirname],axis=0)
    #print('Time')
    #print(data[dirname+'_mean'][-1]/60)
    #print('Convexity')
    #print(DATA[-1])
fig = plt.figure()
ax = fig.add_subplot(111)
ax.hist(DATA)
ax.yaxis.set_ticks_position('both')
ax.xaxis.set_ticks_position('both')
ax.set_ylabel('Number of Regions')
ax.set_xlabel('Convexity')
#ax.grid(which='both')
fig = plt.figure()
ax = fig.add_subplot(111)
ax.scatter(DATA,data)
ax.yaxis.set_ticks_position('both')
ax.xaxis.set_ticks_position('both')
ax.set_ylabel('Build Time (s)')
ax.set_xlabel('Convexity')
plt.show()



import numpy as np
from .basis import *
from .calculate_length import *
from .update import *
from .calculate_radii import *

def set_root(data, boundary, Qterm, gamma, nu,
             Pperm, Pterm, fraction,
             ishomogeneous,isconvex,isdirected,
             start,direction,limit_high,limit_low,volume,
             low=-1,high=0,niter=200):
    """
    Explaination of code
    """
    p0,layer = boundary.pick()
    distance = boundary.volume**(1/3)
    steps = distance//fraction
    path = []
    if not ishomogeneous:
        path,lengths,layers = boundary.path(p0,distance,steps,dive=0)
        p1 = path[-1]
        path = np.array(path)
        data = np.zeros((len(path),data.shape[1]))
    elif start is None and direction is None:
        lengths = [0]
        p1,_ = boundary.pick(homogeneous=True)
        p1 = p1[0]
        if not isconvex:
            attempts = 0
            threshold = (volume)**(1/3)
            tried = set()
            max_attempts = 10
            cell_list_outter = set(boundary.cell_lookup.query_ball_point(p0,threshold+threshold*0.5))
            cell_list_inner  = set(boundary.cell_lookup.query_ball_point(p0,threshold))
            cell_list = list(cell_list_outter.difference(cell_list_inner))
            cell_set = set(cell_list)
            cell_list = list(cell_set - tried)
            while np.sum(lengths) < threshold:
                if attempts > max_attempts or len(cell_list)==0:
                    attempts = 0
                    threshold = threshold*0.9
                    #print(threshold)
                    cell_list_outter = set(boundary.cell_lookup.query_ball_point(p0,threshold+threshold*0.5))
                    cell_list_inner  = set(boundary.cell_lookup.query_ball_point(p0,threshold))
                    cell_list = list(cell_list_outter.difference(cell_list_inner))
                    cell_set = set(cell_list)
                    cell_list = list(cell_set - tried)
                    #ball_size = threshold*1.1
                    while len(cell_list) < 1:
                        ball_size = threshold
                        cell_list_outter = set(boundary.cell_lookup.query_ball_point(p0,ball_size+ball_size*0.5))
                        cell_list_inner  = set(boundary.cell_lookup.query_ball_point(p0,ball_size))
                        cell_list = list(cell_list_outter.difference(cell_list_inner))
                        cell_set = set(cell_list)
                        cell_list = list(cell_set - tried)
                        #if threshold < ((volume)**(1/3)*(0.75**5)):
                        if len(cell_list) < 1000:
                            #cell_list = list(cell_list_outter)
                            max_attmpts = min(len(cell_list),1000)
                        else:
                            print('Outter shell: {}'.format(threshold*1.5))
                            cell_list = list(cell_list_outter)
                            cell_set = set(cell_list)
                            cell_list = list(cell_set - tried)
                            max_attmpts = len(cell_list)//2
                            #cell_list = list(cell_list_outter.difference(cell_list_inner))
                        #cell_set = set(cell_list)
                        #cell_list = list(cell_set - tried)
                        if len(cell_list) > 1:
                            ball_size *= 0.9
                        else:
                            ball_size *= 1.1
                        threshold = ball_size
                    print('Threshold: {}'.format(threshold))
                #p1,_ = boundary.pick(homogeneous=True)
                cell_id = cell_list.pop(0)
                tried.add(cell_id)
                p1,_ = boundary.pick_in_cell(cell_id)
                lengths = [np.linalg.norm(p1-p0)]
                subdivisions = 8
                p1 = p1[0]
                within = True
                for sub in range(1,subdivisions):
                    p_tmp = (p1 - p0)*(sub/subdivisions) + p0
                    if not boundary.within(p_tmp[0],p_tmp[1],p_tmp[2],2):
                        within = False
                        break
                if not within:
                    lengths = [0]
                    attempts += 1
            #threshold = (volume)**(1/3)
            #while not within:
                #boundary.cell_lookup.query_ball_point(p1)
            """
            if not within:
                print('geodesic')
                #i = boundary.tet.grid.find_closest_point(p0)
                #j = boundary.tet.grid.find_closest_point(p1)
                #path,lengths,lines = boundary.get_shortest_path(i,j)
                #path = boundary.tet_pts[path]
                #lengths[-1] = np.linalg.norm(p1-path[-1])
                #lengths[0] = np.linalg.norm(p0-path[0])
                #path = path.tolist()
                #path[0] = p0.tolist()
                #path[-1] = p1.tolist()
                #path = np.array(path)
                path,lengths,res,pf,f,rm = boundary.find_best_path(p0,p1,niter=niter)
                path = np.array(path)
                path,lengths = boundary.resample(path)
                data = np.zeros((len(path)-1,data.shape[1]))
            """
        else:
            while np.sum(lengths) < (volume)**(1/3):
                p1,_ = boundary.pick(homogeneous=True)
                lengths = [np.linalg.norm(p1-p0)]
    else:
        p0 = start
        if limit_high is not None and limit_low is not None:
            distance = limit_high - limit_low
            required_length = limit_low
        elif limit_high is not None and limit_low is None:
            distance = length_high
            required_length = 0
        else:
            distance = volume**(1/3)
            required_length = 0.5*distance
        if direction is None:
            if not boundary.within(start[0],start[1],start[2],2):
                #data = np.zeros((2,data.shape[1]))
                closest_cell = boundary.pv_polydata_surf.find_closest_cell(start)
                faces = boundary.pv_polydata_surf.faces.reshape(len(boundary.pv_polydata_surf.faces)//4,4)
                indices = faces[closest_cell,1:]
                cell_points = boundary.pv_polydata_surf.points[indices]
                #cell_points = np.sum(cell_points,axis=0)/3
                #p1,_ = boundary.pick_in_cell(closest_cell)
                #p1 = p1[0]
                p1 = np.sum(cell_points,axis=0)/3
                """
                lengths = [np.linalg.norm(p1-p0)]
                data[0, 0:3] = p0
                data[0, 3:6] = p1
                basis(data,0)
                data[0, 15] = 1
                data[0, 16] = -1
                data[0, 17] = -1
                data[0, 18] = 0
                data[0, 19] = 1
                data[0, 20] = np.sum(lengths)
                data[0, 22] = Qterm
                data[0, 26] = 0
                data[0, 28] = 1.0
                data[0, 29] = -1
                data[0, -1] = 0
                sub_division_map = [-1]
                sub_division_index = [0]
                update(data, gamma, nu)
                radii(data,Pperm,Pterm)
                """
                p0 = p1 #.flatten()
                lengths = [0]
                #p1,_ = boundary.pick(homogeneous=True)
                #lengths = [np.linalg.norm(p1-p0)]
                if not isconvex:
                    attempts = 0
                    threshold = (volume)**(1/3)
                    max_attempts = 10
                    cell_list_outter = set(boundary.cell_lookup.query_ball_point(p0,threshold+threshold*0.5))
                    cell_list_inner  = set(boundary.cell_lookup.query_ball_point(p0,threshold))
                    cell_list = list(cell_list_outter.difference(cell_list_inner))
                    while np.sum(lengths) < threshold:
                        if attempts > max_attempts or len(cell_list)==0:
                            attempts = 0
                            threshold = threshold*0.9
                            #print(threshold)
                            cell_list_outter = set(boundary.cell_lookup.query_ball_point(p0,threshold+threshold*0.5))
                            cell_list_inner  = set(boundary.cell_lookup.query_ball_point(p0,threshold))
                            cell_list = list(cell_list_outter.difference(cell_list_inner))
                            ball_size = threshold*1.1
                            while len(cell_list) < 1:
                                cell_list_outter = set(boundary.cell_lookup.query_ball_point(p0,ball_size+ball_size*0.5))
                                cell_list_inner  = set(boundary.cell_lookup.query_ball_point(p0,ball_size))
                                if threshold < ((volume)**(1/3)*(0.75**5)):
                                    cell_list = list(cell_list_outter)
                                    max_attmpts = min(len(cell_list),1000)
                                else:
                                    cell_list = list(cell_list_outter.difference(cell_list_inner))
                                ball_size *= 1.1
                        #p1,_ = boundary.pick(homogeneous=True)
                        cell_id = cell_list.pop(0)
                        p1,_ = boundary.pick_in_cell(cell_id)
                        lengths = [np.linalg.norm(p1-p0)]
                        subdivisions = 8
                        p1 = p1[0]
                        within = True
                        for sub in range(1,subdivisions):
                            p_tmp = (p1 - p0)*(sub/subdivisions) + p0
                            if not boundary.within(p_tmp[0],p_tmp[1],p_tmp[2],2):
                                within = False
                                break
                        if not within:
                            lengths = [0]
                            attempts += 1
                else:
                    while np.sum(lengths) < (volume)**(1/3):
                        p1,_ = boundary.pick(homogeneous=True)
                        lengths = [np.linalg.norm(p1-p0)]
                lengths = [np.linalg.norm(start-p0)]
                data[0, 0:3] = start
                data[0, 3:6] = p1
                basis(data,0)
                data[0, 15] = -1
                data[0, 16] = -1
                data[0, 17] = -1
                data[0, 18] = 0
                data[0, 19] = 1
                data[0, 20] = np.sum(lengths)
                data[0, 22] = Qterm
                data[0, 26] = 0
                data[0, 28] = 1.0
                data[0, 29] = -1
                data[0, -1] = 0
                sub_division_map = [-1]
                sub_division_index = [0]
                update(data, gamma, nu)
                radii(data,Pperm,Pterm)
                return data,sub_division_map,sub_division_index
            else:
                p1,_ = boundary.pick(homogeneous=True)
                lengths = [np.linalg.norm(p1-p0)]

        else:
            p1_tmp = start + direction*required_length + direction*distance*np.random.random(1)
            L = np.linalg.norm(p1_tmp - start)
            while not boundary.within(p1_tmp[0],p1_tmp[1],p1_tmp[2],2):
                p1_tmp = p1_tmp - direction*(L/100)
                L = np.linalg.norm(p1_tmp - start)
            p1 = p1_tmp
            lengths = [np.linalg.norm(p1-p0)]
            #print(lengths)
    if len(path) == 0:
        #print('linear root')
        data[0, 0:3] = p0
        data[0, 3:6] = p1
        basis(data,0)
        data[0, 15] = -1
        data[0, 16] = -1
        data[0, 17] = -1
        data[0, 18] = 0
        data[0, 19] = 1
        data[0, 20] = np.sum(lengths)
        data[0, 22] = Qterm
        data[0, 26] = 0
        data[0, 28] = 1.0
        data[0, 29] = -1
        data[0, -1] = 0
        sub_division_map = [-1]
        sub_division_index = [0]
        update(data, gamma, nu)
        radii(data,Pperm,Pterm)
    else:
        sub_division_map = []
        sub_division_index = []
        for i in range(data.shape[0]):
            #if i == 1:
            #    data[i,0:3] = p0
            #    data[i,3:6] = path[i]
            #elif i == data.shape[0]:
            #    data[i,0:3] = path[-1]
            #    data[i,3:6] = p1
            #else:
            #    data[i,0:3] = path[i-1]
            #    data[i,3:6] = path[i]
            data[i,0:3] = path[i]
            data[i,3:6] = path[i+1]
            basis(data,i)
            if i < data.shape[0] - 1:
                data[i,15] = i+1
            else:
                data[i,15] = -1
            data[i,16] = -1
            data[i,17] = i - 1
            data[i,18] = i
            data[i,19] = i+1
            data[i,20] = lengths[i] #length(data,i)
            data[i,22] = Qterm
            data[i,23] = 1.0
            data[i,24] = 0
            data[i,25] = (8*nu/np.pi)*np.sum(lengths[i:])
            data[i,26] = 0
            data[i,27] = np.sum(lengths[i:])
            data[i,28] = 1.0
            data[i,29] = -1
            data[i,-1] = i
            data[i,21] = ((data[0,25]*data[0,22])/(Pperm-Pterm)) ** (1/4)
            sub_division = [-1]
            #sub_division_index.append(len(sub_division_map))
            if i < data.shape[0] - 1:
                sub_division.extend(list(range(i+1,data.shape[0])))
            sub_division_map.extend(sub_division)
            #print(sub_division_map)
    sub_division_index = np.argwhere(np.array(sub_division_map)==-1).flatten().tolist()
    return data,sub_division_map,sub_division_index



import pickle
import glob
import os
import numpy as np
import matplotlib.pyplot as plt

wd = os.getcwd()

nelder_mead = []
powell      = []
bfgs        = []
newton_cg   = []
lbfsgb      = []
tnc         = []
cobyla      = []
slsqp       = []
dogleg      = []
trust_ncg   = []

data = {'Nelder-Mead':{},
        'Powell':{},
        'BFGS':{},
        'Newton-CG':{},
        'L-BFGS-B':{},
        'TNC':{},
        'COBYLA':{},
        'SLSQP':{}}

solvers = ['Nelder-Mead','Powell','BFGS','Newton-CG','L-BFGS-B','TNC',
           'COBYLA','SLSQP']

results = []
for sol in solvers:
    tmp_results = []
    for res in glob.glob(wd+os.sep+sol+os.sep+"*"):
        file = open(res,"rb")
        tmp_results.append(pickle.load(file))
        file.close()
    results.append(tmp_results)

# Get max depth ranges for data
for i in range(len(results)):
    max_depth = 0
    for j in range(len(results[i])):
        max_val = max(results[i][j]['depth'])
        if max_val > max_depth:
            max_depth = max_val
    max_depth = int(max_depth)
    data[solvers[i]]['depth'] = list(range(0,max_depth+1))
    data[solvers[i]]['method_time_bin']   = [[] for k in range(len(data[solvers[i]]['depth']))]
    data[solvers[i]]['brute_time_bin']   = [[] for k in range(len(data[solvers[i]]['depth']))]
    data[solvers[i]]['method_error_bin']   = [[] for k in range(len(data[solvers[i]]['depth']))]
    data[solvers[i]]['brute_error_bin']   = [[] for k in range(len(data[solvers[i]]['depth']))]
    data[solvers[i]]['method_align_bin']   = [[] for k in range(len(data[solvers[i]]['depth']))]
    data[solvers[i]]['brute_align_bin']   = [[] for k in range(len(data[solvers[i]]['depth']))]
# allocate data into correct bins
for i in range(len(results)):
    for j in range(len(results[i])):
        for k in range(len(results[i][j]['depth'])):
            data[solvers[i]]['method_time_bin'][int(results[i][j]['depth'][k])].append(results[i][j]['method_time'][k])
            data[solvers[i]]['brute_time_bin'][int(results[i][j]['depth'][k])].append(results[i][j]['brute_time'][k])
            data[solvers[i]]['method_error_bin'][int(results[i][j]['depth'][k])].append(abs(results[i][j]['truth_value'][k].item() - results[i][j]['method_value'][k])/abs(results[i][j]['truth_value'][k].item()))
            data[solvers[i]]['brute_error_bin'][int(results[i][j]['depth'][k])].append(abs(results[i][j]['truth_value'][k].item() - results[i][j]['brute_value'][k])/abs(results[i][j]['truth_value'][k].item()))
            data[solvers[i]]['method_align_bin'][int(results[i][j]['depth'][k])].append(np.linalg.norm(results[i][j]['truth_x_value'][k] - results[i][j]['method_x_value'][k]))
            data[solvers[i]]['brute_align_bin'][int(results[i][j]['depth'][k])].append(np.linalg.norm(results[i][j]['truth_x_value'][k] - results[i][j]['brute_x_value'][k]))

    data[solvers[i]]['method_time_mean'] = np.array([np.mean(times) for times in data[solvers[i]]['method_time_bin']]).flatten()
    data[solvers[i]]['brute_time_mean'] = np.array([np.mean(times) for times in data[solvers[i]]['brute_time_bin']]).flatten()
    data[solvers[i]]['method_error_mean'] = np.array([np.mean(times) for times in data[solvers[i]]['method_error_bin']]).flatten()
    data[solvers[i]]['brute_error_mean'] = np.array([np.mean(times) for times in data[solvers[i]]['brute_error_bin']]).flatten()
    data[solvers[i]]['method_time_sd'] = np.array([np.std(times) for times in data[solvers[i]]['method_time_bin']]).flatten()
    data[solvers[i]]['brute_time_sd'] = np.array([np.std(times) for times in data[solvers[i]]['brute_time_bin']]).flatten()
    data[solvers[i]]['method_error_sd'] = np.array([np.std(times) for times in data[solvers[i]]['method_error_bin']]).flatten()
    data[solvers[i]]['brute_error_sd'] = np.array([np.std(times) for times in data[solvers[i]]['brute_error_bin']]).flatten()
    data[solvers[i]]['method_align_mean'] = np.array([np.mean(times) for times in data[solvers[i]]['method_align_bin']]).flatten()
    data[solvers[i]]['brute_align_mean'] = np.array([np.mean(times) for times in data[solvers[i]]['brute_align_bin']]).flatten()
    data[solvers[i]]['method_align_sd'] = np.array([np.std(times) for times in data[solvers[i]]['method_align_bin']]).flatten()
    data[solvers[i]]['brute_align_sd'] = np.array([np.std(times) for times in data[solvers[i]]['brute_align_bin']]).flatten()
fig = plt.figure(figsize=[11,9])
nrows = len(solvers)
ncols = 3
idx   = 1
for i in range(len(results)):
    depth = data[solvers[i]]['depth']
    method_times = data[solvers[i]]['method_time_mean']
    brute_times = data[solvers[i]]['brute_time_mean']
    method_times_sd = data[solvers[i]]['method_time_sd']
    brute_times_sd = data[solvers[i]]['brute_time_sd']
    method_error = data[solvers[i]]['method_error_mean']
    brute_error = data[solvers[i]]['brute_error_mean']
    method_error_sd = data[solvers[i]]['method_error_sd']
    brute_error_sd = data[solvers[i]]['brute_error_sd']
    method_align = data[solvers[i]]['method_align_mean']
    brute_align = data[solvers[i]]['brute_align_mean']
    method_align_sd = data[solvers[i]]['method_align_sd']
    brute_align_sd = data[solvers[i]]['brute_align_sd']
    plt.subplot(nrows,ncols,idx)
    plt.plot(depth,method_times,label=solvers[i])
    plt.fill_between(depth,method_times-method_times_sd,method_times+method_times_sd,alpha=0.5)
    plt.plot(depth,brute_times,label="brute")
    plt.fill_between(depth,brute_times-brute_times_sd,brute_times+brute_times_sd,alpha=0.5)
    plt.legend(loc='upper right',prop={'size':6})
    plt.ylim([0,None])
    idx += 1
    plt.subplot(nrows,ncols,idx)
    plt.plot(depth,method_error,label=solvers[i])
    plt.fill_between(depth,method_error-method_error_sd,method_error+method_error_sd,alpha=0.5)
    plt.plot(depth,brute_error,label="brute")
    plt.fill_between(depth,brute_error-brute_error_sd,brute_error+brute_error_sd,alpha=0.5)
    plt.legend(loc='upper right',prop={'size':6})
    plt.ylim([0,None])
    idx += 1
    plt.subplot(nrows,ncols,idx)
    plt.plot(depth,method_align,label=solvers[i])
    plt.fill_between(depth,method_align-method_align_sd,method_align+method_align_sd,alpha=0.5)
    plt.plot(depth,brute_align,label="brute")
    plt.fill_between(depth,brute_align-brute_align_sd,brute_align+brute_align_sd,alpha=0.5)
    plt.legend(loc='upper right',prop={'size':6})
    plt.ylim([0,None])
    idx += 1
plt.subplots_adjust(left=0.1,
                    bottom=0.1,
                    right=0.9,
                    top=0.9,
                    wspace=0.4,
                    hspace=0.4)
fig.text(0.5,0.02,"Bifurcation Depth",ha="center")
fig.text(0.02,0.5,"Time (seconds)",va="center",rotation="vertical")
fig.text(0.328,0.5,"Relative Volume Error",va="center",rotation="vertical")
fig.text(0.625,0.5,"Absolute Alignment Error",va="center",rotation="vertical")
plt.show()



centerline ="""if not terminating:
    outlets = []
    for i,face in enumerate(mesher.get_model_face_ids()):
        if face == walls[0]:
            continue
        elif face == inlet:
            continue
        else:
            outlets.append(face)
    centerline = vmtk.centerlines(mesher.get_model_polydata(),[inlet],outlets,use_face_ids=True)
    centerline_model = modeling.PolyData()
    centerline_model.set_surface(centerline)
    centerline_model.write('{}'+os.sep+'centerlines','vtp')
"""



#!/usr/bin/env python3
"""
This module provides a flexible package for autogenerating vascular networks.

It provides:
  - constrained constructive optimization routines for vasuclar construction
  - implicit surface/volume handling for anatomic and engineered shapes
  - integration with open-source vascular simulation software SimVascular
  - gCode and Stl 3D printing file creation

"""
import re
from pathlib import Path
from setuptools import setup, find_packages
import os
import glob
from tqdm import tqdm
from importlib_metadata import version

here = Path(__file__).parent

long_description = (here / "README.md").read_text("utf8")

#VERSION = re.search(
#    r'__version__ = "(.+?)"', (here / "svcco" / "__init__.py").read_text("utf8")
#).group(1)
VERSION = '0.5.39'

CLASSIFIERS = ['Intended Audience :: Science/Research',
               'License :: OSI Approved :: MIT License',
               'Programming Language :: Python',
               'Programming Language :: Python :: 3.7',
               'Programming Language :: Python :: 3.8',
               'Programming Language :: Python :: 3.9',
               'Topic :: Scientific/Engineering',
               'Operating System :: Microsoft :: Windows',
               'Operating System :: POSIX :: Linux',
               'Operating System :: POSIX',
               'Operating System :: Unix',
               'Operating System :: MacOS']

ALL_FILE_LIST = glob.glob("./**/*.py",recursive=True)
ALL_FILE_LIST = [file for file in ALL_FILE_LIST if "__init__" not in file and "CCO_" not in file]
MODULES = []
"""
for filename in tqdm(ALL_FILE_LIST):
    file = open(filename,'r')
    lines = [line for line in file.readlines() if 'import' in line and line[0] != '#']
    for line in lines:
        words = line.split(" ")
        if words[0] == "import" or words[0] == "from":
            if words[1][0] != ".":
                mod = words[1].split(".")[0].replace('\n','')
                if mod not in MODULES:
                    MODULES.append(mod)
    file.close()

print(MODULES)
INSTALL_REQUIREMENTS = []
for mod in MODULES:
    if mod == 'svcco' or mod == 'importlib_metadata' or mod == 'pip' or mod == 'svZeroDSolver' or mod == 'svzerodsolver':
        continue
    try:
        install_version = version(mod)
        INSTALL_REQUIREMENTS.append(mod+'>='+install_version)
    except:
        pass
print(INSTALL_REQUIREMENTS)
"""
INSTALL_REQUIREMENTS = ['pyvista==0.33.3', 'numpy>=1.21.6', 'setuptools>=62.3.3', 'tqdm>=4.63.0',
                        'matplotlib>=3.3.4', 'vtk>=9.0.1', 'binarytree>=6.5.1', 'memory_profiler>=0.60.0',
                        'networkx>=2.5.1', 'pydotplus>=2.0.2', 'numba>=0.55.1', 'seaborn>=0.11.2',
                        'tetgen>=0.6.0', 'scipy>=1.6.0', 'pymeshfix>=0.15.0', 'geomdl>=5.3.1',
                        'sympy>=1.9', 'plotly>=5.1.0','scikit-image>=0.16.1']

PACKAGES = find_packages(include=["svcco","svcco.*"]) #['svcco']+['svcco.'+ pkg for pkg in find_packages('svcco')]
print(PACKAGES)
OPTIONS  = None
"""
INSTALL_REQUIREMENTS = ['numpy>=1.16.0',
                        'numba>=0.53.0',
                        'scipy>=1.5.0',
                        'pyvista==0.30.1',
                        'matplotlib>=3.3.4',
                        'seaborn>=0.11.0',
                        'sympy>=1.8.0',
                        'tqdm>=4.61.0',
                        'vtk>=9.0.0',
                        'scikit-image>=0.16.1',
                        'pandas>=1.3.0',
                        'plotly>=5.1.0',
                        'pymeshfix>=0.15.0',
                        'tetgen>=0.6.0']
"""
PROJECT_URLS = {}
#PACKAGE_DATA = {}
setup_info = dict(
    name='svcco',
    version=VERSION,
    author='Zachary Sexton',
    author_email='zsexton@stanford.edu',
    url="https://zasexton.github.io/svcco/",
    description='Automated Vascular Generation and CFD Simulation',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    python_requires='>=3.7',
    classifiers=CLASSIFIERS,
    install_requires=INSTALL_REQUIREMENTS,
    packages=PACKAGES
    )

setup(**setup_info)



export_3d_only = """import sv
import vtk
import os

Modeler = sv.modeling.Modeler(sv.modeling.Kernel.POLYDATA)
model = Modeler.read("{}")

mass = vtk.vtkMassProperties()

model.compute_boundary_faces(45)

face_types = []
face_areas = []

face_ids = model.get_face_ids()
caps = model.identify_caps()

inlet_cap_area = -100
inlet_cap = 1

for id in face_ids:
     mass.SetInputData(model.get_face_polydata(id))
     mass.Update()
     SA = mass.GetSurfaceArea()
     face_areas.append(SA)
     if not caps[id-1]:
         face_types.append("Wall")
         print("Wall found")
     else:
         face_types.append("Cap")
         print("Cap found")
         if SA > inlet_cap_area:
             inlet_cap = id
             inlet_cap_area = SA

face_types[inlet_cap-1] = "Inlet"

#---------------------
# REMESHING
#---------------------
elements_on_smallest_face = 20

hmin = min(face_areas)/elements_on_smallest_face
hmax = hmin

optimized_polydata = sv.mesh_utils.remesh(model.get_polydata(),hmin=hmin,hmax=hmax)

model.set_surface(optimized_polydata)
"""



presolver="""if not terminating:
    svpre_file = open('{}'+'/{}.svpre','w+')
    svpre_construction = 'mesh_and_adjncy_vtu mesh-complete/mesh-complete.mesh.vtu\\n'
    svpre_construction += 'set_surface_id_vtp mesh-complete/mesh-complete.exterior.vtp 1\\n'
    resistance = dict([])
    mass = vtk.vtkMassProperties()
    max_surface_area = 0
    total_res = 0
    for i,face in enumerate(mesher.get_model_face_ids()):
        if face == walls[0]:
            continue
        f_poly = mesher.get_face_polydata(face)
        mass.SetInputData(f_poly)
        resistance[face] = mass.GetSurfaceArea()
        total_res += resistance[face]
        if resistance[face] > max_surface_area:
            skip = i + 1
            inlet = face
            max_surface_area = resistance[face]
        svpre_construction += 'set_surface_id_vtp mesh-complete/mesh-surfaces/cap_{{}}.vtp {{}}\\n'.format(face,i+1)
    #total_res -= 1/max_surface_area
    svpre_construction += 'fluid_density {}\\n'
    svpre_construction += 'fluid_viscosity {}\\n'
    svpre_construction += 'initial_pressure {}\\n'
    svpre_construction += 'initial_velocity {} {} {}\\n'
    svpre_construction += 'prescribed_velocities_vtp mesh-complete/mesh-surfaces/cap_{{}}.vtp\\n'.format(inlet)
    svpre_construction += 'bct_analytical_shape {}\\n'
    svpre_construction += 'bct_period {}\\n'
    svpre_construction += 'bct_point_number {}\\n'
    svpre_construction += 'bct_fourier_mode_number {}\\n'
    svpre_construction += 'bct_create mesh-complete/mesh-surfaces/cap_{{}}.vtp {}\\n'.format(inlet)
    svpre_construction += 'bct_write_dat bct.dat\\n'
    svpre_construction += 'bct_write_vtp bct.vtp\\n'
    for i, outlet in enumerate(mesher.get_model_face_ids()):
        if outlet == walls[0]:
            continue
        elif outlet == inlet:
            continue
        svpre_construction += 'pressure_vtp mesh-complete/mesh-surfaces/cap_{{}}.vtp {}\\n'.format(outlet)
    svpre_construction += 'noslip_vtp mesh-complete/walls_combined.vtp\\n'
    svpre_construction += 'write_geombc geombc.dat.1\\n'
    svpre_construction += 'write_restart restart.0.1\\n'
    svpre_file.writelines([svpre_construction])
    svpre_file.close()
"""



import svcco
import pyvista as pv

cube = pv.Cube().triangulate().subdivide(3)

s = svcco.surface()
s.set_data(cube.points,cube.point_normals)
s.solve()
s.build()

# Warm up
t = svcco.tree()
t.convex = True
t.set_boundary(s)
t.set_root()
t.n_add(20)

#t.export(gui=False)



import numpy as np
import numba as nb

@nb.jit(nopython=True)
def N(surface):
    n = surface.shape[0]
    d = surface.shape[1]
    N_matrix = np.zeros((4*n,4))
    for i in range(n):
        N_matrix[i,0] = 1
        for j in range(d):
            N_matrix[i,j+1] = surface[i,j]
    for i in range(n):
        for j in range(d):
            N_matrix[n+i+j*n,j+1] = -1
    return N_matrix
